@@ -1,6 +1,6 @@
 /* ModelicaInternal.c - External functions for Modelica.Utilities
 
-   Copyright (C) 2002-2017, Modelica Association and DLR
+   Copyright (C) 2002-2020, Modelica Association and contributors
    All rights reserved.
 
    Redistribution and use in source and binary forms, with or without
@@ -13,6 +13,10 @@
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.
 
+   3. Neither the name of the copyright holder nor the names of its
+      contributors may be used to endorse or promote products derived from
+      this software without specific prior written permission.
+
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
@@ -25,7 +29,35 @@
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
-/* Release Notes:
+/* Changelog:
+      Nov. 17, 2020: by Thomas Beutlich
+                     Fixed reading files with Unix-style line endings on Windows
+                     for ModelicaInternal_readLine/_readFile (ticket #3631)
+
+      Nov. 11, 2020: by Thomas Beutlich
+                     Added getcwd fallback in ModelicaInternal_fullPathName if
+                     realpath fails for non-existing path (ticket #3660)
+
+      Nov. 13, 2019: by Thomas Beutlich
+                     Utilized blockwise I/O in ModelicaInternal_copyFile
+                     (ticket #3229)
+
+      Oct. 10, 2019: by Thomas Beutlich
+                     Fixed month and year correction in ModelicaInternal_getTime
+                     (ticket #3143)
+
+      Jun. 24, 2019: by Thomas Beutlich
+                     Fixed uninitialized memory and realpath behaviour in
+                     ModelicaInternal_fullPathName (ticket #3003)
+
+      Jun. 28, 2018: by Hans Olsson, Dassault Systemes
+                     Proper error message when out of string memory
+                     in ModelicaInternal_readLine (ticket #2676)
+
+      Oct. 23, 2017: by Thomas Beutlich, ESI ITI GmbH
+                     Utilized non-fatal hash insertion, called by HASH_ADD_KEYPTR in
+                     function CacheFileForReading (ticket #2097)
+
       Apr. 09, 2017: by Thomas Beutlich, ESI ITI GmbH
                      Fixed macOS support of ModelicaInternal_setenv
                      (ticket #2235)
@@ -33,10 +65,6 @@
       Mar. 27, 2017: by Thomas Beutlich, ESI ITI GmbH
                      Replaced localtime by re-entrant function
 
-      Feb. 26, 2017: by Thomas Beutlich, ESI ITI GmbH
-                     Fixed definition of uthash_fatal, called by HASH_ADD_KEYPTR in
-                     function CacheFileForReading (ticket #2097)
-
       Jan. 31, 2017: by Thomas Beutlich, ESI ITI GmbH
                      Fixed WIN32 support of a directory name with a trailing
                      forward/backward slash character in ModelicaInternal_stat
@@ -105,19 +133,8 @@
 */
 
 #include "ModelicaInternal.h"
-#include <string.h>
 #include "ModelicaUtilities.h"
 
-/* The standard way to detect POSIX is to check _POSIX_VERSION,
- * which is defined in <unistd.h>
- */
-#if defined(__unix__) || defined(__linux__) || defined(__APPLE_CC__)
-  #include <unistd.h>
-#endif
-#if !defined(_POSIX_) && defined(_POSIX_VERSION)
-  #define _POSIX_ 1
-#endif
-
 MODELICA_NORETURN static void ModelicaNotExistError(const char* name) MODELICA_NORETURNATTR;
 static void ModelicaNotExistError(const char* name) {
   /* Print error message if a function is not implemented */
@@ -180,10 +197,23 @@ void ModelicaInternal_setenv(_In_z_ const char* name,
     ModelicaNotExistError("ModelicaInternal_setenv"); }
 #else
 
+/* The standard way to detect POSIX is to check _POSIX_VERSION,
+ * which is defined in <unistd.h>
+ */
+#if defined(__unix__) || defined(__linux__) || defined(__APPLE_CC__)
+  #include <unistd.h>
+#endif
+#if !defined(_POSIX_) && defined(_POSIX_VERSION)
+  #define _POSIX_ 1
+#endif
+
+#include "stdint_wrap.h"
+#define HASH_NO_STDINT 1
+#define HASH_NONFATAL_OOM 1
 #include "uthash.h"
-#undef uthash_fatal /* Ensure that nowhere in this file uses uthash_fatal by accident */
 #include "gconstructor.h"
 
+#include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <errno.h>
@@ -204,8 +234,8 @@ void ModelicaInternal_setenv(_In_z_ const char* name,
 
   #if defined(__MINGW32__) || defined(__CYGWIN__) /* MinGW and Cygwin have dirent.h */
     #include <dirent.h>
-  #else /* include the opendir/readdir/closedir implementation for _WIN32 */
-    #include "win32_dirent.c"
+  #else /* include the opendir/readdir/closedir interface for _WIN32 */
+    #include "win32_dirent.h"
   #endif
 
 #elif defined(_POSIX_) || defined(__GNUC__)
@@ -256,6 +286,9 @@ static void ModelicaConvertFromUnixDirectorySeparator(char* string) {
   #define ModelicaConvertFromUnixDirectorySeparator(string) ;
 #endif
 
+static int readLine(_In_ char** buf, _In_ int* bufLen, _In_ FILE* fp) MODELICA_NONNULLATTR;
+  /* Read line (of unknown and arbitrary length) from a text file */
+
 /* --------------------- Modelica_Utilities.Internal --------------------------------- */
 
 void ModelicaInternal_mkdir(_In_z_ const char* directoryName) {
@@ -296,13 +329,13 @@ void ModelicaInternal_rmdir(_In_z_ const char* directoryName) {
 
 static ModelicaFileType Internal_stat(_In_z_ const char* name) {
     /* Inquire type of file */
-    ModelicaFileType type = FileType_NoFile;
+    ModelicaFileType type;
 #if defined(_WIN32)
     struct _stat fileInfo;
     int statReturn = _stat(name, &fileInfo);
     if (0 != statReturn) {
         /* For some reason _stat requires "a:\" and "a:\test1" but fails
-         * on "a:" and "a:\test1\", repectively. It could be handled in the
+         * on "a:" and "a:\test1\", respectively. It could be handled in the
          * Modelica code, but seems better to have it here.
          */
         const char* firstSlash = strpbrk(name, "/\\");
@@ -357,6 +390,8 @@ static ModelicaFileType Internal_stat(_In_z_ const char* name) {
     else {
         type = FileType_SpecialFile;
     }
+#else
+    type = FileType_NoFile;
 #endif
     return type;
 }
@@ -391,17 +426,11 @@ void ModelicaInternal_removeFile(_In_z_ const char* file) {
 void ModelicaInternal_copyFile(_In_z_ const char* oldFile,
                                _In_z_ const char* newFile) {
     /* Copy file */
-#ifdef _WIN32
     const char* modeOld = "rb";
     const char* modeNew = "wb";
-#else
-    const char* modeOld = "r";
-    const char* modeNew = "w";
-#endif
     FILE* fpOld;
     FILE* fpNew;
     ModelicaFileType type;
-    int c;
 
     /* Check file existence */
     type = Internal_stat(oldFile);
@@ -438,8 +467,18 @@ void ModelicaInternal_copyFile(_In_z_ const char* oldFile,
             oldFile, newFile, strerror(errno));
         return;
     }
-    while ( (c = getc(fpOld)) != EOF ) {
-        putc(c, fpNew);
+    {
+        size_t len;
+        char buf[BUFSIZ] = {'\0'};
+
+        while ( (len = fread(buf, sizeof(char), BUFSIZ, fpOld)) > 0 ) {
+            if ( len != fwrite(buf, sizeof(char), len, fpNew) ) {
+                fclose(fpOld);
+                fclose(fpNew);
+                ModelicaFormatError("Error writing to file \"%s\".", newFile);
+                return;
+            }
+        }
     }
     fclose(fpOld);
     fclose(fpNew);
@@ -514,7 +553,7 @@ void ModelicaInternal_readDirectory(_In_z_ const char* directory, int nFiles,
              directory, iFiles, nFiles);
     }
     else if ( closedir(pdir) != 0 ) {
-        ModelicaFormatError("Not possible to get file names of \"%s\":\n",
+        ModelicaFormatError("Not possible to get file names of \"%s\":\n%s",
             directory, strerror(errno));
     }
 
@@ -566,15 +605,10 @@ Modelica_ERROR:
 _Ret_z_ const char* ModelicaInternal_fullPathName(_In_z_ const char* name) {
     /* Get full path name of file or directory */
 
-#if defined(_WIN32) || (_BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED || (_POSIX_VERSION >= 200112L))
+#if defined(_WIN32)
     char* fullName;
     char localbuf[BUFFER_LENGTH];
-#if (_BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED || _POSIX_VERSION >= 200112L)
-    /* realpath availability: 4.4BSD, POSIX.1-2001. Using the behaviour of NULL: POSIX.1-2008 */
-    char* tempName = realpath(name, localbuf);
-#else
     char* tempName = _fullpath(localbuf, name, sizeof(localbuf));
-#endif
     if (tempName == NULL) {
         ModelicaFormatError("Not possible to construct full path name of \"%s\"\n%s",
             name, strerror(errno));
@@ -583,28 +617,58 @@ _Ret_z_ const char* ModelicaInternal_fullPathName(_In_z_ const char* name) {
     fullName = ModelicaAllocateString(strlen(tempName));
     strcpy(fullName, tempName);
     ModelicaConvertToUnixDirectorySeparator(fullName);
+    return fullName;
+#elif (_BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED || _POSIX_VERSION >= 200112L)
+    char* fullName;
+    char localbuf[BUFFER_LENGTH];
+    size_t len;
+    /* realpath availability: 4.4BSD, POSIX.1-2001. Using the behaviour of NULL: POSIX.1-2008 */
+    char* tempName = realpath(name, localbuf);
+    if (tempName == NULL) {
+        goto FALLBACK_getcwd;
+    }
+    fullName = ModelicaAllocateString(strlen(tempName) + 1);
+    strcpy(fullName, tempName);
+    ModelicaConvertToUnixDirectorySeparator(fullName);
+    /* Retain trailing slash to match _fullpath behaviour */
+    len = strlen(name);
+    if (len > 0 && '/' == name[len - 1]) {
+        strcat(fullName, "/");
+    }
+    return fullName;
 #elif defined(_POSIX_)
     char* fullName;
     char localbuf[BUFFER_LENGTH];
-    /* No such system call in _POSIX_ available (except realpath above) */
-    char* cwd = getcwd(localbuf, sizeof(localbuf));
-    if (cwd == NULL) {
-        ModelicaFormatError("Not possible to get current working directory:\n%s",
-            strerror(errno));
-    }
-    fullName = ModelicaAllocateString(strlen(cwd) + strlen(name) + 1);
-    if (name[0] != '/') {
-        /* Any name beginning with "/" is regarded as already being a full path. */
-        strcpy(fullName, cwd);
-        strcat(fullName, "/");
-    }
-    strcat(fullName, name);
 #else
     char* fullName = "";
     ModelicaNotExistError("ModelicaInternal_fullPathName");
+    return fullName;
 #endif
 
+#if (_BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED || _POSIX_VERSION >= 200112L)
+FALLBACK_getcwd:
+#endif
+#if (_BSD_SOURCE || _XOPEN_SOURCE >= 500 || _XOPEN_SOURCE && _XOPEN_SOURCE_EXTENDED || _POSIX_VERSION >= 200112L || _POSIX_)
+    {
+        /* No such system call in _POSIX_ available (except realpath for existing paths) */
+        char* cwd = getcwd(localbuf, sizeof(localbuf));
+        if (cwd == NULL) {
+            ModelicaFormatError("Not possible to get current working directory:\n%s",
+                strerror(errno));
+        }
+        fullName = ModelicaAllocateString(strlen(cwd) + strlen(name) + 1);
+        if (name[0] != '/') {
+            /* Any name beginning with "/" is regarded as already being a full path. */
+            strcpy(fullName, cwd);
+            strcat(fullName, "/");
+        }
+        else {
+            fullName[0] = '\0';
+        }
+        strcat(fullName, name);
+    }
     return fullName;
+#endif
 }
 
 _Ret_z_ const char* ModelicaInternal_temporaryFileName(void) {
@@ -629,12 +693,14 @@ _Ret_z_ const char* ModelicaInternal_temporaryFileName(void) {
 typedef struct FileCache {
     char* fileName; /* Key = File name */
     FILE* fp /* File pointer */;
-    int line;
+    char* buf;
+    int bufLen;
+    int lineNumber;
     UT_hash_handle hh; /* Hashable structure */
 } FileCache;
 
 static FileCache* fileCache = NULL;
-#if defined(_POSIX_)
+#if defined(_POSIX_) && !defined(NO_MUTEX)
 #include <pthread.h>
 #if defined(G_HAS_CONSTRUCTORS)
 static pthread_mutex_t m;
@@ -662,17 +728,17 @@ static pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;
 #include <windows.h>
 static CRITICAL_SECTION cs;
 #ifdef G_DEFINE_CONSTRUCTOR_NEEDS_PRAGMA
-#pragma G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(initializeCS)
+#pragma G_DEFINE_CONSTRUCTOR_PRAGMA_ARGS(ModelicaInternal_initializeCS)
 #endif
-G_DEFINE_CONSTRUCTOR(initializeCS)
-static void initializeCS(void) {
+G_DEFINE_CONSTRUCTOR(ModelicaInternal_initializeCS)
+static void ModelicaInternal_initializeCS(void) {
     InitializeCriticalSection(&cs);
 }
 #ifdef G_DEFINE_DESTRUCTOR_NEEDS_PRAGMA
-#pragma G_DEFINE_DESTRUCTOR_PRAGMA_ARGS(deleteCS)
+#pragma G_DEFINE_DESTRUCTOR_PRAGMA_ARGS(ModelicaInternal_deleteCS)
 #endif
-G_DEFINE_DESTRUCTOR(deleteCS)
-static void deleteCS(void) {
+G_DEFINE_DESTRUCTOR(ModelicaInternal_deleteCS)
+static void ModelicaInternal_deleteCS(void) {
     DeleteCriticalSection(&cs);
 }
 #define MUTEX_LOCK() EnterCriticalSection(&cs)
@@ -682,14 +748,13 @@ static void deleteCS(void) {
 #define MUTEX_UNLOCK()
 #endif
 
-static void CacheFileForReading(FILE* fp, const char* fileName, int line) {
-#define uthash_fatal(msg) do { \
-    MUTEX_UNLOCK(); \
-    ModelicaFormatMessage("Error in uthash: %s\n" \
-        "Hash table for file cache may be left in corrupt state.\n", msg); \
-    return; \
-} while (0)
+#if !defined(LINE_BUFFER_LENGTH)
+#define LINE_BUFFER_LENGTH (64)
+#endif
+
+static void CacheFileForReading(FILE* fp, const char* fileName, int lineNumber, char* buf, int bufLen) {
     FileCache* fv;
+    size_t len;
     if (fileName == NULL) {
         /* Do not add, close file */
         if (fp != NULL) {
@@ -697,22 +762,31 @@ static void CacheFileForReading(FILE* fp, const char* fileName, int line) {
         }
         return;
     }
+    len = strlen(fileName);
     MUTEX_LOCK();
-    HASH_FIND(hh, fileCache, fileName, (unsigned)strlen(fileName), fv);
+    HASH_FIND(hh, fileCache, fileName, (unsigned)len, fv);
     if (fv != NULL) {
         fv->fp = fp;
-        fv->line = line;
+        fv->lineNumber = lineNumber;
+        fv->buf = buf;
+        fv->bufLen = bufLen;
     }
     else {
         fv = (FileCache*)malloc(sizeof(FileCache));
         if (fv != NULL) {
-            char* key = (char*)malloc((strlen(fileName) + 1)*sizeof(char));
+            char* key = (char*)malloc((len + 1)*sizeof(char));
             if (key != NULL) {
                 strcpy(key, fileName);
                 fv->fileName = key;
                 fv->fp = fp;
-                fv->line = line;
-                HASH_ADD_KEYPTR(hh, fileCache, key, (unsigned)strlen(key), fv);
+                fv->lineNumber = lineNumber;
+                fv->buf = buf;
+                fv->bufLen = bufLen;
+                HASH_ADD_KEYPTR(hh, fileCache, key, (unsigned)len, fv);
+                if (NULL == fv->hh.tbl) {
+                   free(key);
+                   free(fv);
+                }
             }
             else {
                 free(fv);
@@ -720,17 +794,18 @@ static void CacheFileForReading(FILE* fp, const char* fileName, int line) {
         }
     }
     MUTEX_UNLOCK();
-#undef uthash_fatal
 }
 
 static void CloseCachedFile(const char* fileName) {
     FileCache* fv;
+    size_t len = strlen(fileName);
     MUTEX_LOCK();
-    HASH_FIND(hh, fileCache, fileName, (unsigned)strlen(fileName), fv);
+    HASH_FIND(hh, fileCache, fileName, (unsigned)len, fv);
     if (fv != NULL) {
         if (fv->fp != NULL) {
             fclose(fv->fp);
         }
+        free(fv->buf);
         free(fv->fileName);
         HASH_DEL(fileCache, fv);
         free(fv);
@@ -738,31 +813,40 @@ static void CloseCachedFile(const char* fileName) {
     MUTEX_UNLOCK();
 }
 
-static FILE* ModelicaStreams_openFileForReading(const char* fileName, int line) {
+static FILE* ModelicaStreams_openFileForReading(const char* fileName, int lineNumber, int* lineNumberOffset, char** buf, int* bufLen) {
     /* Open text file for reading */
     FILE* fp;
-    int c = 1;
     FileCache* fv;
+    size_t len = strlen(fileName);
+    *lineNumberOffset = 0;
+    *buf = NULL;
+    *bufLen = LINE_BUFFER_LENGTH;
     MUTEX_LOCK();
-    HASH_FIND(hh, fileCache, fileName, (unsigned)strlen(fileName), fv);
+    HASH_FIND(hh, fileCache, fileName, (unsigned)len, fv);
     /* Open file */
     if (fv != NULL) {
         /* Cached value */
         if (fv->fp != NULL) {
-            if (line != 0 && line >= fv->line) {
-                line -= fv->line;
+            if (lineNumber != 0 && lineNumber >= fv->lineNumber - 1) {
+                *lineNumberOffset = fv->lineNumber;
                 fp = fv->fp;
+                *buf = fv->buf;
+                *bufLen = fv->bufLen;
             }
             else {
                 if ( fseek(fv->fp, 0L, SEEK_SET) == 0 ) {
                     fp = fv->fp;
+                    *buf = fv->buf;
+                    *bufLen = fv->bufLen;
                 }
                 else {
                     fclose(fv->fp);
                     fp = NULL;
+                    free(fv->buf);
                 }
             }
             fv->fp = NULL;
+            fv->buf = NULL;
         }
         else {
             fp = NULL;
@@ -779,13 +863,6 @@ static FILE* ModelicaStreams_openFileForReading(const char* fileName, int line)
                 "%s\n", fileName, strerror(errno));
         }
     }
-    while ( line != 0 && c != EOF ) {
-        c = fgetc(fp);
-        while ( c != '\n' && c != EOF ) {
-            c = fgetc(fp);
-        }
-        line--;
-    }
     return fp;
 }
 
@@ -799,7 +876,7 @@ static FILE* ModelicaStreams_openFileForWriting(const char* fileName) {
     FILE* fp;
 
     /* Check fileName */
-    if ( strlen(fileName) == 0 ) {
+    if ( fileName[0] == '\0' ) {
         ModelicaError("fileName is an empty string.\n"
             "Opening of file is aborted\n");
     }
@@ -814,6 +891,43 @@ static FILE* ModelicaStreams_openFileForWriting(const char* fileName) {
     return fp;
 }
 
+static int readLine(_In_ char** buf, _In_ int* bufLen, _In_ FILE* fp) {
+    char* offset;
+    int oldBufLen;
+
+    if (fgets(*buf, *bufLen, fp) == NULL) {
+        return EOF;
+    }
+
+    do {
+        char* p;
+        char* tmp;
+
+        if ((p = strchr(*buf, '\n')) != NULL) {
+            *p = '\0';
+            return 0;
+        }
+        if ((p = memchr(*buf, 0, (size_t)(*bufLen - 1))) != NULL) {
+            return 1;
+        }
+
+        oldBufLen = *bufLen;
+        *bufLen *= 2;
+        tmp = (char*)realloc(*buf, (size_t)*bufLen);
+        if (NULL == tmp) {
+            fclose(fp);
+            free(*buf);
+            ModelicaError("Memory allocation error\n");
+            return 1;
+        }
+        *buf = tmp;
+        offset = &((*buf)[oldBufLen - 1]);
+
+    } while (fgets(offset, oldBufLen + 1, fp));
+
+    return 0;
+}
+
 /* --------------------- Modelica_Utilities.Streams ----------------------------------- */
 
 void ModelicaInternal_print(_In_z_ const char* string,
@@ -844,12 +958,15 @@ Modelica_ERROR2:
 
 int ModelicaInternal_countLines(_In_z_ const char* fileName) {
     /* Get number of lines of a file */
+    int lineNumberOffset;
+    int bufLen;
+    char* buf;
     int c;
     int nLines = 0;
     int start_of_line = 1;
     /* If true, next character starts a new line. */
 
-    FILE* fp = ModelicaStreams_openFileForReading(fileName, 0);
+    FILE* fp = ModelicaStreams_openFileForReading(fileName, 0, &lineNumberOffset, &buf, &bufLen);
 
     /* Count number of lines */
     while ((c = fgetc(fp)) != EOF) {
@@ -868,119 +985,83 @@ int ModelicaInternal_countLines(_In_z_ const char* fileName) {
 void ModelicaInternal_readFile(_In_z_ const char* fileName,
                                _Out_ const char** string, size_t nLines) {
     /* Read file into string vector string[nLines] */
-    FILE* fp = ModelicaStreams_openFileForReading(fileName, 0);
+    int lineNumberOffset;
+    int bufLen;
+    char* buf;
+    FILE* fp = ModelicaStreams_openFileForReading(fileName, 0, &lineNumberOffset, &buf, &bufLen);
     char* line;
-    size_t iLines;
-    size_t nc;
-    char localbuf[200]; /* To avoid fseek */
+    size_t iLines = 1;
+
+    if (buf == NULL) {
+        buf = (char*)calloc(bufLen, sizeof(char));
+        if (buf == NULL) {
+            goto Modelica_OOM_ERROR1;
+        }
+    }
 
     /* Read data from file */
-    iLines = 1;
-    while ( iLines <= nLines ) {
-        /* Determine length of next line */
-        long offset = ftell(fp);
-        size_t lineLen = 0;
-        int c = fgetc(fp);
-        int c2 = c;
-        while ( c != '\n' && c != EOF ) {
-            if (lineLen < sizeof(localbuf)) {
-                localbuf[lineLen] = (char)c;
-            }
-            lineLen++;
-            c2 = c;
-            c = fgetc(fp);
-        }
+    while (iLines <= nLines) {
+        readLine(&buf, &bufLen, fp);
 
-        if ( lineLen > 0 && c2 == '\r' ) {
-            lineLen--;
-        }
-        /* Allocate storage for next line */
-        line = ModelicaAllocateStringWithErrorReturn(lineLen);
+        line = ModelicaAllocateStringWithErrorReturn(strlen(buf));
         if ( line == NULL ) {
-            fclose(fp);
-            ModelicaFormatError("Not enough memory to allocate string for reading line %i from file\n"
-                "\"%s\".\n"
-                "(this file contains %i lines)\n", iLines, fileName, nLines);
+            goto Modelica_OOM_ERROR1;
         }
 
-        /* Read next line */
-        if (lineLen<=sizeof(localbuf)) {
-            memcpy(line, localbuf, lineLen);
-        }
-        else {
-            if ( fseek(fp, offset, SEEK_SET != 0) ) {
-                fclose(fp);
-                ModelicaFormatError("Error when reading line %i from file\n\"%s\":\n"
-                    "%s\n", iLines, fileName, strerror(errno));
-            }
-            nc = ( iLines < nLines ? lineLen+1 : lineLen);
-            if ( fread(line, sizeof(char), nc, fp) != nc ) {
-                fclose(fp);
-                ModelicaFormatError("Error when reading line %i from file\n\"%s\"\n",
-                    iLines, fileName);
-            }
-        }
-        line[lineLen] = '\0';
-        string[iLines-1] = line;
+        strcpy(line, buf);
+        string[iLines - 1] = line;
         iLines++;
     }
     fclose(fp);
+    free(buf);
+    return;
+
+    /* Out-of-memory error */
+Modelica_OOM_ERROR1:
+    fclose(fp);
+    free(buf);
+    ModelicaFormatError("Error when reading line %lu from file \"%s\":\n"
+        "Not enough memory to allocate string for reading line.",
+        (unsigned long)iLines, fileName);
 }
 
 _Ret_z_ const char* ModelicaInternal_readLine(_In_z_ const char* fileName,
                                       int lineNumber, _Out_ int* endOfFile) {
     /* Read line lineNumber from file fileName */
-    FILE* fp = ModelicaStreams_openFileForReading(fileName, lineNumber - 1);
+    int lineNumberOffset;
+    int bufLen;
+    char* buf;
+    FILE* fp;
     char* line;
-    int c, c2;
-    size_t lineLen;
-    long offset;
-    char localbuf[200]; /* To avoid fseek */
+    int iLine;
+
+    fp = ModelicaStreams_openFileForReading(fileName, lineNumber - 1, &lineNumberOffset, &buf, &bufLen);
 
     if (feof(fp)) {
         goto END_OF_FILE;
     }
 
-    /* Determine length of line lineNumber */
-    offset  = ftell(fp);
-    lineLen = 0;
-    c = fgetc(fp);
-    c2 = c;
-    while ( c != '\n' && c != EOF ) {
-        if (lineLen < sizeof(localbuf)) {
-            localbuf[lineLen] = (char)c;
+    if (buf == NULL) {
+        buf = (char*)calloc(bufLen, sizeof(char));
+        if (buf == NULL) {
+            goto Modelica_OOM_ERROR2;
         }
-        lineLen++;
-        c2 = c;
-        c = fgetc(fp);
-    }
-    if ( lineLen == 0 && c == EOF ) {
-        goto END_OF_FILE;
     }
 
-    /* Read line lineNumber */
-    if ( lineLen > 0 && c2 == '\r') {
-        lineLen--;
-    }
-    line = ModelicaAllocateStringWithErrorReturn(lineLen);
-    if ( line == NULL ) {
-        goto Modelica_ERROR3;
+    for (iLine = 0; iLine < lineNumber - lineNumberOffset; iLine++) {
+        int readError = readLine(&buf, &bufLen, fp);
+        if (readError == EOF && iLine == lineNumber - lineNumberOffset - 1) {
+            goto END_OF_FILE;
+        }
     }
 
-    if (lineLen <= sizeof(localbuf)) {
-        memcpy(line, localbuf, lineLen);
+    line = ModelicaAllocateStringWithErrorReturn(strlen(buf));
+    if (line == NULL) {
+        goto Modelica_OOM_ERROR2;
     }
-    else {
-        if ( fseek(fp, offset, SEEK_SET) != 0 ) {
-            goto Modelica_ERROR3;
-        }
-        if ( fread(line, sizeof(char), lineLen, fp) != lineLen ) {
-            goto Modelica_ERROR3;
-        }
-        fgetc(fp); /* Read the EOF/new-line. */
-    }
-    CacheFileForReading(fp, fileName, lineNumber);
-    line[lineLen] = '\0';
+
+    strcpy(line, buf);
+    CacheFileForReading(fp, fileName, lineNumber, buf, bufLen);
     *endOfFile = 0;
     return line;
 
@@ -990,13 +1071,14 @@ END_OF_FILE:
     CloseCachedFile(fileName);
     *endOfFile = 1;
     line = ModelicaAllocateString(0);
+    line[0] = '\0';
     return line;
 
-Modelica_ERROR3:
+Modelica_OOM_ERROR2:
     fclose(fp);
     CloseCachedFile(fileName);
-    ModelicaFormatError("Error when reading line %i from file\n\"%s\":\n%s",
-        lineNumber, fileName, strerror(errno));
+    ModelicaFormatError("Error when reading line %i from file \"%s\":\n"
+        "Not enough memory to allocate string for reading line.", lineNumber, fileName);
     return "";
 }
 
@@ -1252,14 +1334,14 @@ void ModelicaInternal_getTime(_Out_ int* ms, _Out_ int* sec, _Out_ int* min, _Ou
     struct tm tres;
 #endif
 
-    time( &calendarTime );               /* Retrieve sec time */
+    time(&calendarTime);                        /* Retrieve sec time */
 #if defined(_POSIX_)
     tlocal = localtime_r(&calendarTime, &tres); /* Time fields in local time zone */
 #elif defined(_MSC_VER) && _MSC_VER >= 1400
     localtime_s(&tres, &calendarTime);          /* Time fields in local time zone */
     tlocal = &tres;
 #else
-    tlocal = localtime( &calendarTime );        /* Time fields in local time zone */
+    tlocal = localtime(&calendarTime);          /* Time fields in local time zone */
 #endif
 
     /* Get millisecond resolution depending on platform */
@@ -1275,9 +1357,7 @@ void ModelicaInternal_getTime(_Out_ int* ms, _Out_ int* sec, _Out_ int* min, _Ou
 #else
         _ftime( &timebuffer );                    /* Retrieve ms time */
 #endif
-        ms0 = (int)(timebuffer.millitm);          /* Convert unsigned int to int */
-        tlocal->tm_mon  = tlocal->tm_mon + 1;     /* Correct for month starting at 1 */
-        tlocal->tm_year = tlocal->tm_year + 1900; /* Correct for 4-digit year */
+        ms0 = (int)(timebuffer.millitm);        /* Convert unsigned int to int */
     }
 #else
     {
@@ -1293,7 +1373,7 @@ void ModelicaInternal_getTime(_Out_ int* ms, _Out_ int* sec, _Out_ int* min, _Ou
     *min = tlocal->tm_min;
     *hour = tlocal->tm_hour;
     *mday = tlocal->tm_mday;
-    *mon = tlocal->tm_mon;
-    *year = tlocal->tm_year;
+    *mon = 1 + tlocal->tm_mon;      /* Correct for month starting at 1 */
+    *year = 1900 + tlocal->tm_year; /* Correct for 4-digit year */
 #endif
 }
