@@ -113,7 +113,7 @@ def ask_omc(question, opt=None, parsed=True):
 
     try:
         if parsed:
-            res = omc.execute(expression)
+            res = omc.sendExpression(expression)
             omc.clearOMParserResult()
         else:
             res = omc.sendExpression(expression, parsed=False)
@@ -773,7 +773,6 @@ def getSvgFromGraphics(dwg, graphics, minX, maxY, includeInvisibleText, transfor
                 shape.push(x_12, y_12, x_n, y_n, x_n, y_n)
             else:
                 shape = dwg.polygon([getCoordinates([x, y], graphics, minX, maxY, transformation, coordinateSystem) for (x, y) in graphics['points']])
-            shape.fill('none', opacity=0)
         else:
             logger.error('Not handled: {0}'.format(graphics))
             return None
