@@ -60,6 +60,7 @@
 
 #include "util/omc_error.h"
 #include "util/omc_file.h"
+#include "util/omc_numbers.h"
 #include "simulation_data.h"
 #include "openmodelica_func.h"
 #include "meta/meta_modelica.h"
@@ -86,6 +87,7 @@
 #include "simulation/solver/initialization/initialization.h"
 #include "simulation/solver/dae_mode.h"
 #include "dataReconciliation/dataReconciliation.h"
+#include "util/parallel_helper.h"
 
 #ifdef _OMC_QSS_LIB
   #include "solver_qss/solver_qss.h"
@@ -186,9 +188,16 @@ void setGlobalVerboseLevel(int argc, char**argv)
     }while(pos != string::npos);
   }
 
-  /* print LOG_SOTI if LOG_INIT is enabled */
-  if(useStream[LOG_INIT])
+  /* print LOG_INIT and LOG_SOTI if LOG_INIT_V is active */
+  if(useStream[LOG_INIT_V] == 1)
+  {
+    useStream[LOG_INIT] = 1;
     useStream[LOG_SOTI] = 1;
+  }
+
+  /* print LOG_INIT_HOMOTOPY if LOG_INIT is active */
+  if(useStream[LOG_INIT] == 1)
+    useStream[LOG_INIT_HOMOTOPY] = 1;
 
   /* print LOG_STATS if LOG_SOLVER if active */
   if(useStream[LOG_SOLVER_V] == 1)
@@ -226,6 +235,51 @@ void setGlobalVerboseLevel(int argc, char**argv)
   delete flags;
 }
 
+
+/* Read value of flag lv_time to set time interval in which loggin is active */
+void setGlobalLoggingTime(int argc, char**argv, SIMULATION_INFO *simulationInfo)
+{
+  const char *flagStr = omc_flagValue[FLAG_LV_TIME];
+  const string *flags = flagStr ? new string(flagStr) : NULL;
+  char *endptr;
+  const char *secondPart;
+  double loggingStartTime, loggingStopTime;
+
+  /* Check if lv_time flag is given */
+  if (flagStr==NULL || *flagStr=='\0')
+  {
+    /* default activated --> Log everything*/
+    simulationInfo->useLoggingTime = 0;
+    return;
+  }
+
+  /* Parse flagStr */
+  loggingStartTime = om_strtod(flagStr, &endptr);
+  endptr = endptr+1;
+  secondPart = endptr;
+  loggingStopTime = om_strtod(secondPart, &endptr);
+  if (*endptr)
+  {
+    throwStreamPrint(NULL, "Simulation flag %s expects two real numbers, seperated by a comata. Got: %s", FLAG_NAME[FLAG_LV_TIME], flagStr);
+  }
+
+  /* Check flag input */
+  if (loggingStartTime > loggingStopTime)
+  {
+    throwStreamPrint(NULL, "Simulation flag %s expects first number to be smaller then second number. Got: %s", FLAG_NAME[FLAG_LV_TIME], flagStr);
+  }
+
+  /* Save logging time */
+  simulationInfo->useLoggingTime = 1;
+  simulationInfo->loggingTimeRecord[0] = loggingStartTime;
+  simulationInfo->loggingTimeRecord[1] = loggingStopTime;
+  infoStreamPrint(LOG_STDOUT, 0, "Time dependent logging enabled. Activate loggin in intervall [%f, %f]", simulationInfo->loggingTimeRecord[0], simulationInfo->loggingTimeRecord[1]);
+
+  /* Deactivate Logging */
+  deactivateLogging();
+}
+
+
 static void readFlag(int *flag, int max, const char *value, const char *flagName, const char **names, const char **desc)
 {
   int i;
@@ -255,7 +309,7 @@ static double getFlagReal(enum _FLAG flag, double res)
   if (flagStr==NULL || *flagStr=='\0') {
     return res;
   }
-  res = strtod(flagStr, &endptr);
+  res = om_strtod(flagStr, &endptr);
   if (*endptr) {
     throwStreamPrint(NULL, "Simulation flag %s expects a real number, got: %s", FLAG_NAME[flag], flagStr);
   }
@@ -431,7 +485,7 @@ int startNonInteractiveSimulation(int argc, char**argv, DATA* data, threadData_t
     } else {
       data->simulationInfo->stopTime = atof(lintime);
     }
-    infoStreamPrint(LOG_STDOUT, 0, "Linearization will performed at point of time: %f", data->simulationInfo->stopTime);
+    infoStreamPrint(LOG_STDOUT, 0, "Linearization will be performed at point of time: %f", data->simulationInfo->stopTime);
   }
 
   /* set delta x for linearization */
@@ -481,7 +535,6 @@ int startNonInteractiveSimulation(int argc, char**argv, DATA* data, threadData_t
   string init_file = "";
   string init_time_string = "";
   double init_time = 0.0;
-  string init_lambda_steps_string = "";
   string outputVariablesAtEnd = "";
   int cpuTime = omc_flag[FLAG_CPU];
 
@@ -489,15 +542,30 @@ int startNonInteractiveSimulation(int argc, char**argv, DATA* data, threadData_t
     init_initMethod = omc_flagValue[FLAG_IIM];
   }
   if(omc_flag[FLAG_IIF]) {
-    init_file = omc_flagValue[FLAG_IIF];
+    if (omc_flag[FLAG_INPUT_PATH]) {
+      const char *tmp_filename;
+      if (0 > GC_asprintf(&tmp_filename, "%s/%s", omc_flagValue[FLAG_INPUT_PATH], omc_flagValue[FLAG_IIF])) {
+        throwStreamPrint(NULL, "simulation_runtime.cpp: Error: can not allocate memory.");
+      }
+      init_file = tmp_filename;
+    }
+    else {
+      init_file = omc_flagValue[FLAG_IIF];
+    }
   }
   if(omc_flag[FLAG_IIT]) {
     init_time_string = omc_flagValue[FLAG_IIT];
     init_time = atof(init_time_string.c_str());
   }
   if(omc_flag[FLAG_ILS]) {
-    init_lambda_steps_string = omc_flagValue[FLAG_ILS];
-    init_lambda_steps = atoi(init_lambda_steps_string.c_str());
+    init_lambda_steps = atoi(omc_flagValue[FLAG_ILS]);
+    if(init_lambda_steps <= 0) {
+      init_lambda_steps = 0;
+      infoStreamPrint(LOG_STDOUT, 0, "Number of lambda steps set to 0. Homotopy is disabled.");
+    }
+    else {
+      infoStreamPrint(LOG_STDOUT, 0, "Number of lambda steps for homotopy approach changed to %d", init_lambda_steps);
+    }
   }
   if(omc_flag[FLAG_MAX_BISECTION_ITERATIONS]) {
     maxBisectionIterations = atoi(omc_flagValue[FLAG_MAX_BISECTION_ITERATIONS]);
@@ -511,25 +579,34 @@ int startNonInteractiveSimulation(int argc, char**argv, DATA* data, threadData_t
     outputVariablesAtEnd = omc_flagValue[FLAG_OUTPUT];
   }
 
+  /* Check if logging should be enabled */
+  if ((data->simulationInfo->useLoggingTime == 1) && (data->simulationInfo->startTime >= data->simulationInfo->loggingTimeRecord[0])) {
+    reactivateLogging();
+  }
+
   retVal = callSolver(data, threadData, init_initMethod, init_file, init_time, outputVariablesAtEnd, cpuTime, argv[0]);
 
+  /* Check if logging should be disabled */
+  if (data->simulationInfo->useLoggingTime == 1) {
+    deactivateLogging();
+  }
+
   if (omc_flag[FLAG_ALARM]) {
     alarm(0);
   }
 
-  if(0 == retVal && omc_flag[FLAG_DATA_RECONCILE])
+  if (omc_flag[FLAG_DATA_RECONCILE])
   {
-	infoStreamPrint(LOG_STDOUT, 0, "DataReconciliation Starting!");
-	infoStreamPrint(LOG_STDOUT, 0, "%s", data->modelData->modelName);
-	retVal = dataReconciliation(data, threadData);
-	infoStreamPrint(LOG_STDOUT, 0, "DataReconciliation Completed!");
+    infoStreamPrint(LOG_STDOUT, 0, "DataReconciliation Starting!");
+    infoStreamPrint(LOG_STDOUT, 0, "%s", data->modelData->modelName);
+    retVal = dataReconciliation(data, threadData, retVal);
+    infoStreamPrint(LOG_STDOUT, 0, "DataReconciliation Completed!");
   }
 
   if(0 == retVal && create_linearmodel) {
     rt_tick(SIM_TIMER_JACOBIAN);
     retVal = linearize(data, threadData);
     rt_accumulate(SIM_TIMER_JACOBIAN);
-    infoStreamPrint(LOG_STDOUT, 0, "Linear model is created!");
   }
 
   /* Use the saved state of measure_time_flag.
@@ -664,6 +741,7 @@ static int callSolver(DATA* simData, threadData_t *threadData, string init_initM
       )
   {
     solverID = S_EULER;
+    infoStreamPrint(LOG_SOLVER, 0, "No states present, continuing without ODE solver.");
     if (compiledInDAEMode)
     {
       simData->callback->functionDAE = evaluateDAEResiduals_wrapperEventUpdate;
@@ -709,7 +787,29 @@ int initRuntimeAndSimulation(int argc, char**argv, DATA *data, threadData_t *thr
   int i;
   initDumpSystem();
 
-  if(setLogFormat(argc, argv) || helpFlagSet(argc, argv) || checkCommandLineArguments(argc, argv))
+  int checkArgumentsRes = checkCommandLineArguments(argc, argv);
+
+#ifndef NO_INTERACTIVE_DEPENDENCY
+  if(omc_flag[FLAG_PORT]) {
+    std::istringstream stream(omc_flagValue[FLAG_PORT]);
+    int port;
+    stream >> port;
+    sim_communication_port_open = 1;
+    sim_communication_port_open &= sim_communication_port.create();
+    sim_communication_port_open &= sim_communication_port.connect("127.0.0.1", port);
+  }
+#endif
+
+  int logFormatResult = setLogFormat(argc, argv);
+
+#ifndef NO_INTERACTIVE_DEPENDENCY
+  if (isXMLTCP && !sim_communication_port_open) {
+    errorStreamPrint(LOG_STDOUT, 0, "xmltcp log format requires a TCP-port to be passed (and successfully open)");
+    EXIT(1);
+  }
+#endif
+
+  if(logFormatResult || helpFlagSet(argc, argv) || checkArgumentsRes)
   {
     infoStreamPrint(LOG_STDOUT, 1, "usage: %s", argv[0]);
 
@@ -819,6 +919,7 @@ int initRuntimeAndSimulation(int argc, char**argv, DATA *data, threadData_t *thr
   }
 
   setGlobalVerboseLevel(argc, argv);
+  setGlobalLoggingTime(argc, argv, data->simulationInfo);
   initializeDataStruc(data, threadData);
   if(!data)
   {
@@ -940,6 +1041,36 @@ int initRuntimeAndSimulation(int argc, char**argv, DATA *data, threadData_t *thr
   data->simulationInfo->minStepSize = 4.0 * DBL_EPSILON * fmax(fabs(data->simulationInfo->startTime),fabs(data->simulationInfo->stopTime));
   rt_accumulate(SIM_TIMER_INIT_XML);
 
+  /* Set the maximum number of threads prior to any allocation w.r.t.
+   * linear systems and Jacobians in order to avoid memory leaks.
+   */
+#ifdef USE_PARJAC
+  int num_threads = omc_get_max_threads();
+  if (omc_flag[FLAG_JACOBIAN_THREADS]) {
+    int num_threads_tmp = atoi(omc_flagValue[FLAG_JACOBIAN_THREADS]);
+    infoStreamPrint(LOG_STDOUT, 0,
+         "Number of threads passed via -jacobianThreads: %d",
+         num_threads_tmp);
+    if (0 >= num_threads_tmp) {
+      warningStreamPrint(LOG_STDOUT, 0,
+          "Number of desired OpenMP threads for parallel Jacobian evaluation is <= 0.");
+      warningStreamPrint(LOG_STDOUT, 0, "Use omp_get_max_threads().");
+    } else {
+      num_threads = num_threads_tmp;
+    }
+  }
+  omp_set_num_threads(num_threads);
+
+  infoStreamPrint(LOG_STDOUT, 0,
+      "Number of OpenMP threads for parallel Jacobian evaluation: %d",
+      omc_get_max_threads());
+#else
+  if (omc_flag[FLAG_JACOBIAN_THREADS]) {
+      warningStreamPrint(LOG_STDOUT, 0,
+          "Simulation flag jacobianThreads not available. Make sure you have configured omc with \"--enable-parjac\" and build with a compiler supporting OpenMP.");
+  }
+#endif
+
   /* initialize static data of mixed/linear/non-linear system solvers */
   initializeMixedSystems(data, threadData);
   initializeLinearSystems(data, threadData);
@@ -949,22 +1080,10 @@ int initRuntimeAndSimulation(int argc, char**argv, DATA *data, threadData_t *thr
 
 #ifndef NO_INTERACTIVE_DEPENDENCY
   if(omc_flag[FLAG_PORT]) {
-    std::istringstream stream(omc_flagValue[FLAG_PORT]);
-    int port;
-    stream >> port;
-    sim_communication_port_open = 1;
-    sim_communication_port_open &= sim_communication_port.create();
-    sim_communication_port_open &= sim_communication_port.connect("127.0.0.1", port);
-
     if(0 != strcmp("ia", data->simulationInfo->outputFormat)) {
       communicateStatus("Starting", 0.0, data->simulationInfo->startTime, 0);
     }
   }
-
-  if (isXMLTCP && !sim_communication_port_open) {
-    errorStreamPrint(LOG_STDOUT, 0, "xmltcp log format requires a TCP-port to be passed (and successfully open)");
-    EXIT(1);
-  }
 #endif
   // ppriv - NO_INTERACTIVE_DEPENDENCY - for simpler debugging in Visual Studio
 
