@@ -34,27 +34,35 @@
 #include <float.h>
 #include <math.h>
 
-#include "simulation_data.h"
-#include "openmodelica_func.h"
-#include "util/omc_error.h"
-#include "util/varinfo.h"
+#include "../../openmodelica.h"
+#include "../../simulation_data.h"
+#include "../../openmodelica_func.h"
+#include "../../util/omc_error.h"
+#include "../../util/varinfo.h"
 #include "model_help.h"
-#include "simulation/options.h"
-#include "simulation/simulation_info_json.h"
-#include "util/omc_msvc.h" /* for freaking round! */
+#include "../options.h"
+#include "../simulation_info_json.h"
+#include "../../util/omc_msvc.h" /* for freaking round! */
 #include "nonlinearSystem.h"
 #include "linearSystem.h"
 #include "mixedSystem.h"
 #include "delay.h"
 #include "epsilon.h"
-#include "simulation/solver/stateset.h"
-#include "meta/meta_modelica.h"
+#include "fmi_events.h"
+#include "stateset.h"
+#include "spatialDistribution.h"
+#include "../../meta/meta_modelica.h"
+
+#ifdef USE_PARJAC
+  #include <omp.h>
+#endif
 
 int maxEventIterations = 20;
 double linearSparseSolverMaxDensity = 0.2;
-int linearSparseSolverMinSize = 4001;
+int linearSparseSolverMinSize = DEFAULT_FLAG_LSS_MIN_SIZE;
 double nonlinearSparseSolverMaxDensity = 0.2;
-int nonlinearSparseSolverMinSize = 10001;
+int nonlinearSparseSolverMinSize = DEFAULT_FLAG_NLS_MIN_SIZE;
+double maxStepFactor = 1e12;
 double newtonXTol = 1e-12;
 double newtonFTol = 1e-12;
 double steadyStateTol = 1e-3;
@@ -63,6 +71,19 @@ int compiledInDAEMode = 0;
 int compiledWithSymSolver = 0;
 double numericalDifferentiationDeltaXlinearize = 1e-8;
 double numericalDifferentiationDeltaXsolver = 1e-8;
+double homAdaptBend = 0.5;
+double homHEps = 1e-5;
+int homMaxLambdaSteps = 0;
+int homMaxNewtonSteps = 20;
+int homMaxTries = 10;
+double homTauDecreasingFactor = 10.0;
+double homTauDecreasingFactorPredictor = 2.0;
+double homTauIncreasingFactor = 2.0;
+double homTauIncreasingThreshold = 10.0;
+double homTauMax = 10.0;
+double homTauMin = 1e-4;
+double homTauStart = 0.2;
+int homBacktraceStrategy = 1;
 
 static double tolZC;
 
@@ -91,7 +112,7 @@ void updateDiscreteSystem(DATA *data, threadData_t *threadData)
   debugStreamPrint(LOG_EVENTS_V, 0, "updated discrete System");
 
   relationChanged = checkRelations(data);
-  discreteChanged = data->callback->checkForDiscreteChanges(data, threadData);
+  discreteChanged = checkForDiscreteChanges(data, threadData);
   while(discreteChanged || data->simulationInfo->needToIterate || relationChanged)
   {
     if(data->simulationInfo->needToIterate) {
@@ -118,7 +139,7 @@ void updateDiscreteSystem(DATA *data, threadData_t *threadData)
     }
 
     relationChanged = checkRelations(data);
-    discreteChanged = data->callback->checkForDiscreteChanges(data, threadData);
+    discreteChanged = checkForDiscreteChanges(data, threadData);
   }
   storeRelations(data);
 
@@ -350,31 +371,42 @@ void printParameters(DATA *data, int stream)
   TRACE_POP
 }
 
-/*! \fn printSparseStructure
+/**
+ * @brief Prints sparse structure.
  *
- *  prints sparse structure of jacobian A
- *
- *  \param [in]  [sparsePattern]
- *  \param [in]  [sizeRow]
- *  \param [in]  [sizeCol]
- *  \param [in]  [stream]
+ * Use to print e.g. sparse Jacobian matrix.
+ * Only prints if stream is active and sparse pattern is non NULL and of size > 0.
  *
+ * @param [in]  sparsePattern   Matrix to print.
+ * @param [in]  sizeRows        Number of rows of matrix.
+ * @param [in]  sizeCols        Number of columns of matrix.
+ * @param [in]  stream          Steam to print to.
+ * @param [in]  name            Name of matrix.
  */
 void printSparseStructure(SPARSE_PATTERN *sparsePattern, int sizeRows, int sizeCols, int stream, const char* name)
 {
+  /* Variables */
   unsigned int row, col, i, j;
-  /* Will crash with a static size array */
-  char *buffer = NULL;
+  char *buffer;
 
   if (!ACTIVE_STREAM(stream))
+  {
     return;
+  }
+
+  /* Catch empty sparsePattern */
+  if (sparsePattern == NULL || sizeRows <= 0 || sizeCols <= 0)
+  {
+    infoStreamPrint(stream, 0, "No sparse strucutre available for \"%s\".", name);
+    return;
+  }
 
   buffer = (char*)omc_alloc_interface.malloc(sizeof(char)* 2*sizeCols + 4);
 
-  infoStreamPrint(stream, 1, "sparse structure of %s [size: %ux%u]", name, sizeRows, sizeCols);
+  infoStreamPrint(stream, 1, "Sparse structure of %s [size: %ux%u]", name, sizeRows, sizeCols);
   infoStreamPrint(stream, 0, "%u nonzero elements", sparsePattern->numberOfNoneZeros);
 
-  infoStreamPrint(stream, 1, "transposed sparse structure (rows: states)");
+  infoStreamPrint(stream, 1, "Transposed sparse structure (rows: states)");
   i=0;
   for(row=0; row < sizeRows; row++)
   {
@@ -828,15 +860,16 @@ void storeRelations(DATA* data)
   TRACE_POP
 }
 
-/*! \fn getNextSampleTimeFMU
+/**
+ * @brief Get time of next sample event if one is defined.
  *
- *  function return next sample time.
+ * Function returns 0 if a time is defined and -1 otherwise.
  *
- *  \param [in]  [data]
- *
- *  \author wbraun
+ * @param data                  Data
+ * @param nextSampleEvent       On output time of next sample event.
+ * @return int                  1 if a sample event is defined, 0 otherwise
  */
-double getNextSampleTimeFMU(DATA *data)
+int getNextSampleTimeFMU(DATA *data, double *nextSampleEvent)
 {
   TRACE_PUSH
 
@@ -844,11 +877,12 @@ double getNextSampleTimeFMU(DATA *data)
   {
     infoStreamPrint(LOG_EVENTS, 0, "Next event time = %f", data->simulationInfo->nextSampleEvent);
     TRACE_POP
-    return data->simulationInfo->nextSampleEvent;
+    *nextSampleEvent = data->simulationInfo->nextSampleEvent;
+    return 1 /* TRUE */;
   }
 
   TRACE_POP
-  return -1;
+  return 0 /* FALSE */;
 }
 
 /*! \fn initializeDataStruc
@@ -879,14 +913,14 @@ void initializeDataStruc(DATA *data, threadData_t *threadData)
     tmpSimData.timeValue = 0;
     /* buffer for all variable values */
     tmpSimData.realVars = (modelica_real*) calloc(data->modelData->nVariablesReal, sizeof(modelica_real));
-    assertStreamPrint(threadData, 0 != tmpSimData.realVars, "out of memory");
+    assertStreamPrint(threadData, 0 == data->modelData->nVariablesReal || 0 != tmpSimData.realVars, "out of memory");
     tmpSimData.integerVars = (modelica_integer*) calloc(data->modelData->nVariablesInteger, sizeof(modelica_integer));
-    assertStreamPrint(threadData, 0 != tmpSimData.integerVars, "out of memory");
+    assertStreamPrint(threadData, 0 == data->modelData->nVariablesInteger || 0 != tmpSimData.integerVars, "out of memory");
     tmpSimData.booleanVars = (modelica_boolean*) calloc(data->modelData->nVariablesBoolean, sizeof(modelica_boolean));
-    assertStreamPrint(threadData, 0 != tmpSimData.booleanVars, "out of memory");
+    assertStreamPrint(threadData, 0 == data->modelData->nVariablesBoolean || 0 != tmpSimData.booleanVars, "out of memory");
 #if !defined(OMC_NVAR_STRING) || OMC_NVAR_STRING>0
     tmpSimData.stringVars = (modelica_string*) omc_alloc_interface.malloc_uncollectable(data->modelData->nVariablesString * sizeof(modelica_string));
-    assertStreamPrint(threadData, 0 != tmpSimData.stringVars, "out of memory");
+    assertStreamPrint(threadData, 0 == data->modelData->nVariablesString || 0 != tmpSimData.stringVars, "out of memory");
 #endif
     appendRingData(data->simulationData, &tmpSimData);
   }
@@ -919,17 +953,20 @@ void initializeDataStruc(DATA *data, threadData_t *threadData)
   data->modelData->clocksInfo = (CLOCK_INFO*) omc_alloc_interface.malloc_uncollectable(data->modelData->nClocks * sizeof(CLOCK_INFO));
   data->modelData->subClocksInfo = (SUBCLOCK_INFO*) omc_alloc_interface.malloc_uncollectable(data->modelData->nSubClocks * sizeof(SUBCLOCK_INFO));
   data->simulationInfo->clocksData = (CLOCK_DATA*) calloc(data->modelData->nClocks, sizeof(CLOCK_DATA));
+  data->simulationInfo->spatialDistributionData = allocSpatialDistribution(data->modelData->nSpatialDistributions);
+  data->simulationInfo->intvlTimers = NULL;
 
   /* set default solvers for algebraic loops */
 #if !defined(OMC_MINIMAL_RUNTIME)
-  data->simulationInfo->nlsMethod = NLS_HYBRID;
+  data->simulationInfo->nlsMethod = NLS_MIXED;
 #else
   data->simulationInfo->nlsMethod = NLS_HOMOTOPY;
 #endif
-  data->simulationInfo->lsMethod = LS_LAPACK;
-  data->simulationInfo->lssMethod = LS_UMFPACK;
+  data->simulationInfo->nlsLinearSolver = NLS_LS_DEFAULT;
+  data->simulationInfo->lsMethod = LS_DEFAULT;
+  data->simulationInfo->lssMethod = LSS_DEFAULT;
   data->simulationInfo->mixedMethod = MIXED_SEARCH;
-  data->simulationInfo->newtonStrategy = NEWTON_PURE;
+  data->simulationInfo->newtonStrategy = NEWTON_DAMPED2;
   data->simulationInfo->nlsCsvInfomation = 0;
   data->simulationInfo->currentContext = CONTEXT_ALGEBRAIC;
   data->simulationInfo->jacobianEvals = data->modelData->nStates;
@@ -968,6 +1005,8 @@ void initializeDataStruc(DATA *data, threadData_t *threadData)
   /* buffer for inputs and outputs values */
   data->simulationInfo->inputVars = (modelica_real*) calloc(data->modelData->nInputVars, sizeof(modelica_real));
   data->simulationInfo->outputVars = (modelica_real*) calloc(data->modelData->nOutputVars, sizeof(modelica_real));
+  data->simulationInfo->setcVars = (modelica_real*) calloc(data->modelData->nSetcVars, sizeof(modelica_real));
+  data->simulationInfo->datainputVars = (modelica_real*) calloc(data->modelData->ndataReconVars, sizeof(modelica_real));
 
 #if !defined(OMC_NUM_MIXED_SYSTEMS) || OMC_NUM_MIXED_SYSTEMS>0
   /* buffer for mixed systems */
@@ -1029,7 +1068,7 @@ void initializeDataStruc(DATA *data, threadData_t *threadData)
   data->simulationInfo->extObjs = NULL;
   data->simulationInfo->extObjs = (void**) calloc(data->modelData->nExtObjs, sizeof(void*));
 
-  assertStreamPrint(threadData, 0 != data->simulationInfo->extObjs, "error allocating external objects");
+  assertStreamPrint(threadData, 0 == data->modelData->nExtObjs || 0 != data->simulationInfo->extObjs, "error allocating external objects");
 
 #if !defined(OMC_MINIMAL_LOGGING)
   /* initial chattering info */
@@ -1067,7 +1106,7 @@ void initializeDataStruc(DATA *data, threadData_t *threadData)
   /* initial delay */
 #if !defined(OMC_NDELAY_EXPRESSIONS) || OMC_NDELAY_EXPRESSIONS>0
   data->simulationInfo->delayStructure = (RINGBUFFER**)malloc(data->modelData->nDelayExpressions * sizeof(RINGBUFFER*));
-  assertStreamPrint(threadData, 0 != data->simulationInfo->delayStructure, "out of memory");
+  assertStreamPrint(threadData, 0 == data->modelData->nDelayExpressions || 0 != data->simulationInfo->delayStructure, "out of memory");
 
   for(i=0; i<data->modelData->nDelayExpressions; i++)
     data->simulationInfo->delayStructure[i] = allocRingBuffer(1024, sizeof(TIME_AND_VALUE));
@@ -1145,6 +1184,10 @@ void deInitializeDataStruc(DATA *data)
 
   omc_alloc_interface.free_uncollectable(data->modelData->clocksInfo);
   omc_alloc_interface.free_uncollectable(data->modelData->subClocksInfo);
+  free(data->simulationInfo->clocksData);
+
+  freeSpatialDistribution(data->simulationInfo->spatialDistributionData, data->modelData->nSpatialDistributions);
+  free(data->simulationInfo->spatialDistributionData);
 
   /* free simulationInfo arrays */
   free(data->simulationInfo->zeroCrossings);
@@ -1154,6 +1197,7 @@ void deInitializeDataStruc(DATA *data)
   free(data->simulationInfo->relationsPre);
   free(data->simulationInfo->storedRelations);
   free(data->simulationInfo->zeroCrossingIndex);
+  free(data->simulationInfo->mathEventsValuePre);
 
   /* free buffer for old state variables */
   free(data->simulationInfo->realVarsOld);
@@ -1173,11 +1217,6 @@ void deInitializeDataStruc(DATA *data)
   free(data->simulationInfo->booleanParameter);
   omc_alloc_interface.free_uncollectable(data->simulationInfo->stringParameter);
 
-  if (data->modelData->nStateSets) {
-    /* free buffer for state sets */
-    omc_alloc_interface.free_uncollectable(data->simulationInfo->stateSetData);
-  }
-
   if (data->modelData->nMixedSystems) {
     /* free buffer of mixed systems */
     omc_alloc_interface.free_uncollectable(data->simulationInfo->mixedSystemData);
@@ -1200,9 +1239,16 @@ void deInitializeDataStruc(DATA *data)
   /* free buffer for state sets */
   omc_alloc_interface.free_uncollectable(data->simulationInfo->daeModeData);
 
+  /* buffer for inline Data */
+  free(data->simulationInfo->inlineData->algVars);
+  free(data->simulationInfo->inlineData->algOldVars);
+  omc_alloc_interface.free_uncollectable(data->simulationInfo->inlineData);
+
   /* free inputs and output */
   free(data->simulationInfo->inputVars);
   free(data->simulationInfo->outputVars);
+  free(data->simulationInfo->setcVars);
+  free(data->simulationInfo->datainputVars);
 
   /* free external objects buffer */
   free(data->simulationInfo->extObjs);
@@ -1221,6 +1267,10 @@ void deInitializeDataStruc(DATA *data)
   /* free stateset data */
   freeStateSetData(data);
 #endif
+  if (data->modelData->nStateSets) {
+    /* free buffer for state sets */
+    omc_alloc_interface.free_uncollectable(data->simulationInfo->stateSetData);
+  }
 
   /* free parameter sensitivities */
   if (omc_flag[FLAG_IDAS])
@@ -1230,6 +1280,9 @@ void deInitializeDataStruc(DATA *data)
     FREE_VARS(nSensitivityVars, realSensitivityData)
   }
 
+  /* Free model info xml data */
+  modelInfoDeinit(&(data->modelData->modelDataXml));
+
   TRACE_POP
 }
 
@@ -1373,8 +1426,8 @@ modelica_integer _event_mod_integer(modelica_integer x1, modelica_integer x2, mo
     data->simulationInfo->mathEventsValuePre[index] = (modelica_real)x1;
     data->simulationInfo->mathEventsValuePre[index+1] = (modelica_real)x2;
   }
-
-  return x1 - (x1 / x2) * x2;
+  modelica_integer tmp = x1 % x2;
+  return ((x2 > 0 && tmp < 0) || (x2 < 0 && tmp > 0)) ? (tmp + x2) : tmp;
 }
 
 /*! \fn _event_mod_real
@@ -1460,6 +1513,7 @@ const char *context_string[CONTEXT_MAX] = {
  "context algebraic evaluation",
  "context event search",
  "context jacobian evaluation",
+ "context symbolica jacobian evaluation"
 };
 
 /*! \fn setContext
@@ -1473,8 +1527,10 @@ const char *context_string[CONTEXT_MAX] = {
 void setContext(DATA* data, double* currentTime, int currentContext){
   data->simulationInfo->currentContextOld =  data->simulationInfo->currentContext;
   data->simulationInfo->currentContext =  currentContext;
-  infoStreamPrint(LOG_SOLVER_CONTEXT, 0, "+++ Set context %s +++ at time %f", context_string[data->simulationInfo->currentContext], *currentTime);
-  if (currentContext == CONTEXT_JACOBIAN){
+  infoStreamPrint(LOG_SOLVER_CONTEXT, 0, "+++ Set context %s +++ at time %f", context_string[currentContext], *currentTime);
+  if (currentContext == CONTEXT_JACOBIAN ||
+      currentContext == CONTEXT_SYM_JACOBIAN)
+  {
     data->simulationInfo->currentJacobianEval = 0;
   }
 }
@@ -1486,9 +1542,12 @@ void setContext(DATA* data, double* currentTime, int currentContext){
  * Increase Jacobian column context in simulation info object
  */
 void increaseJacContext(DATA* data){
-  if (data->simulationInfo->currentContext == CONTEXT_JACOBIAN){
+  int currentContext = data->simulationInfo->currentContext;
+  if (currentContext == CONTEXT_JACOBIAN ||
+      currentContext == CONTEXT_SYM_JACOBIAN)
+  {
     data->simulationInfo->currentJacobianEval++;
-    infoStreamPrint(LOG_SOLVER_CONTEXT, 0, "+++ Increase Jacobian column context %s +++ to %d", context_string[data->simulationInfo->currentContext], data->simulationInfo->currentJacobianEval);
+    infoStreamPrint(LOG_SOLVER_CONTEXT, 0, "+++ Increase Jacobian column context %s +++ to %d", context_string[currentContext], data->simulationInfo->currentJacobianEval);
   }
 }
 
