@@ -819,19 +819,43 @@ int solveNLS(DATA *data, threadData_t *threadData, int sysNumber)
   case NLS_HYBRID:
     solverData = nonlinsys->solverData;
     nonlinsys->solverData = solverData->ordinaryData;
+    /* try */
+    #ifndef OMC_EMCC
+      MMC_TRY_INTERNAL(simulationJumpBuffer)
+    #endif
     success = solveHybrd(data, threadData, sysNumber);
+    /*catch */
+    #ifndef OMC_EMCC
+      MMC_CATCH_INTERNAL(simulationJumpBuffer)
+    #endif
     nonlinsys->solverData = solverData;
     break;
   case NLS_KINSOL:
     solverData = nonlinsys->solverData;
     nonlinsys->solverData = solverData->ordinaryData;
+    /* try */
+    #ifndef OMC_EMCC
+      MMC_TRY_INTERNAL(simulationJumpBuffer)
+    #endif
     success = nlsKinsolSolve(data, threadData, sysNumber);
+    /*catch */
+    #ifndef OMC_EMCC
+      MMC_CATCH_INTERNAL(simulationJumpBuffer)
+    #endif
     nonlinsys->solverData = solverData;
     break;
   case NLS_NEWTON:
     solverData = nonlinsys->solverData;
     nonlinsys->solverData = solverData->ordinaryData;
+    /* try */
+    #ifndef OMC_EMCC
+      MMC_TRY_INTERNAL(simulationJumpBuffer)
+    #endif
     success = solveNewton(data, threadData, sysNumber);
+    /*catch */
+    #ifndef OMC_EMCC
+      MMC_CATCH_INTERNAL(simulationJumpBuffer)
+    #endif
     /* check if solution process was successful, if not use alternative tearing set if available (dynamic tearing)*/
     if (!success && casualTearingSet){
       debugString(LOG_DT, "Solving the casual tearing set failed! Now the strict tearing set is used.");
@@ -849,6 +873,10 @@ int solveNLS(DATA *data, threadData_t *threadData, int sysNumber)
     mixedSolverData = nonlinsys->solverData;
     nonlinsys->solverData = mixedSolverData->newtonHomotopyData;
 
+    /* try */
+    #ifndef OMC_EMCC
+      MMC_TRY_INTERNAL(simulationJumpBuffer)
+    #endif
     /* try to solve the system if it is the strict set, only try to solve the casual set if the constraints are satisfied */
     if ((!casualTearingSet) || constraintsSatisfied)
       success = solveHomotopy(data, threadData, sysNumber);
@@ -872,6 +900,10 @@ int solveNLS(DATA *data, threadData_t *threadData, int sysNumber)
       nonlinsys->getIterationVars(data, nonlinsys->nlsx);
     }
 
+    /*catch */
+    #ifndef OMC_EMCC
+      MMC_CATCH_INTERNAL(simulationJumpBuffer)
+    #endif
     nonlinsys->solverData = mixedSolverData;
     break;
 #endif
@@ -949,10 +981,12 @@ int solve_nonlinear_system(DATA *data, threadData_t *threadData, int sysNumber)
   int j;
   int nlsLs;
   int kinsol = 0;
+  int res;
   struct dataSolver *solverData;
   struct dataMixedSolver *mixedSolverData;
   char buffer[4096];
   FILE *pFile = NULL;
+  double originalLambda = data->simulationInfo->lambda;
 
 #if !defined(OMC_MINIMAL_RUNTIME)
   kinsol = (data->simulationInfo->nlsMethod == NLS_KINSOL);
@@ -1160,7 +1194,9 @@ int solve_nonlinear_system(DATA *data, threadData_t *threadData, int sysNumber)
     );
   }
 #endif
-  return check_nonlinear_solution(data, 1, sysNumber);
+  res = check_nonlinear_solution(data, 1, sysNumber);
+  data->simulationInfo->lambda = originalLambda;
+  return res;
 }
 
 /*! \fn check_nonlinear_solutions
