@@ -31,15 +31,16 @@
 
 #include "Model.h"
 
+#include "Component.h"
 #include "CSVWriter.h"
 #include "Flags.h"
 #include "MATWriter.h"
+#include "OMSFileSystem.h"
+#include "OMSString.h"
 #include "Scope.h"
 #include "ssd/Tags.h"
 #include "System.h"
-#include "Component.h"
 
-#include <OMSFileSystem.h>
 #include <minizip.h>
 #include <thread>
 
@@ -118,36 +119,57 @@ oms_status_enu_t oms::Model::rename(const oms::ComRef& cref)
     return logError(std::string(cref) + " is not a valid ident");
 
   this->cref = cref;
+
+  if (system)
+  {
+    system->renameConnectors();
+    for (const auto & it:system->getSubSystems())
+      it.second->renameConnectors();
+  }
+
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::loadSnapshot(const char* snapshot)
+oms_status_enu_t oms::Model::rename(const ComRef& cref, const ComRef& newCref)
 {
+  if (!newCref.isValidIdent())
+    return logError(std::string(newCref) + " is not a valid ident");
+
+  if (!system)
+    return logError("Model \"" + std::string(getCref()) + "\" does not contain any system");
+
+  // renaming the model
+  if (cref.isEmpty())
+    return this->rename(newCref);
+
+  // renaming any subcomponent
+  oms::ComRef tail(cref);
+  oms::ComRef front = tail.pop_front();
+
+  if (system->getCref() == front)
+    return system->rename(tail, newCref);
+
+  return logError("Model \"" + std::string(getCref()) + "\" does not contain system \"" + std::string(front) + "\"");
+}
+
+oms_status_enu_t oms::Model::loadSnapshot(const pugi::xml_node& node)
+{
+  // This method will not change the name of the model.
+  // If a renaming is requested then it will happen in Scope::loadSnapshot.
+
   if (!validState(oms_modelState_virgin))
-    return logError_ModelInWrongState(this);
-
-  pugi::xml_document doc;
-  pugi::xml_parse_result result = doc.load(snapshot);
-  if (!result)
-    return logError("loading snapshot failed (" + std::string(result.description()) + ")");
-
-  const pugi::xml_node node = doc.document_element(); // ssd:SystemStructureDescription
-
-  ComRef new_cref = ComRef(node.attribute("name").as_string());
-  std::string ssdVersion = node.attribute("version").as_string();
-
-  if (new_cref != getCref())
-    return logError("this API cannot be used to rename a model");
-
-  if (ssdVersion != "Draft20180219" && ssdVersion != "1.0")
-    logWarning("Unknown SSD version: " + ssdVersion);
+    return logError_ModelInWrongState(getCref());
 
   System* old_root_system = system;
   system = NULL;
 
+  Snapshot snapshot; // this is a temporary workaroud, loadSnapshot will be removed later
+  snapshot.importResourceNode("SystemStructure.ssd", node);
+  //snapshot.debugPrintAll();
+
   bool old_copyResources = copyResources();
   copyResources(false);
-  oms_status_enu_t status = importFromSSD(node);
+  oms_status_enu_t status = importFromSnapshot(snapshot);
   copyResources(old_copyResources);
 
   if (oms_status_ok != status)
@@ -165,26 +187,38 @@ oms_status_enu_t oms::Model::loadSnapshot(const char* snapshot)
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::importSnapshot(const char* snapshot)
+oms_status_enu_t oms::Model::importSnapshot(const char* snapshot_, char** newCref)
 {
   if (!validState(oms_modelState_virgin))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
-  pugi::xml_document doc;
-  pugi::xml_parse_result result = doc.load(snapshot);
-  if (!result)
-    return logError("loading snapshot failed (" + std::string(result.description()) + ")");
+  Snapshot snapshot;
+  snapshot.import(snapshot_);
+  //snapshot.debugPrintAll();
 
-  snapShot = doc.document_element(); // oms:snapshot
+  // get the new root cref from the snapshot, this should be done here
+  // at the top before importing the full snapshot, as the new cref
+  // will be overwritten
+  new_root_cref = snapshot.getRootCref();
+
+  if (snapshot.isPartialSnapshot())
+  {
+    char* fullsnapshot;
+    // get full snapshot
+    exportSnapshot("", &fullsnapshot);
+    snapshot.importPartialSnapshot(fullsnapshot);
+    free(fullsnapshot);
+  }
 
   // get ssd:SystemStructureDescription
-  pugi::xml_node ssdNode = snapShot.child(oms::ssp::Version1_0::ssd_file).child(oms::ssp::Draft20180219::ssd::system_structure_description);
+  pugi::xml_node ssdNode = snapshot.getResourceNode("SystemStructure.ssd");
 
   ComRef new_cref = ComRef(ssdNode.attribute("name").as_string());
   std::string ssdVersion = ssdNode.attribute("version").as_string();
 
-  if (new_cref != getCref())
-    return logError("this API cannot be used to rename a model");
+  // check new_cref exists in scope
+  if (new_cref != getCref() && Scope::GetInstance().getModel(new_cref))
+    return logError("Renaming the model \"" + std::string(getCref()) + "\" to \"" + std::string(new_cref) + "\" failed because another model with the same name already exists in the scope.");
 
   if (ssdVersion != "Draft20180219" && ssdVersion != "1.0")
     logWarning("Unknown SSD version: " + ssdVersion);
@@ -194,7 +228,7 @@ oms_status_enu_t oms::Model::importSnapshot(const char* snapshot)
 
   bool old_copyResources = copyResources();
   copyResources(false);
-  oms_status_enu_t status = importFromSSD(ssdNode);
+  oms_status_enu_t status = importFromSnapshot(snapshot);
   copyResources(old_copyResources);
 
   if (oms_status_ok != status)
@@ -209,6 +243,13 @@ oms_status_enu_t oms::Model::importSnapshot(const char* snapshot)
     old_root_system = NULL;
   }
 
+  if (newCref)
+    *newCref = (char*)new_root_cref.c_str();
+
+  // rename model
+  if (new_cref != getCref())
+    Scope::GetInstance().renameModel(getCref(), new_cref);
+
   return oms_status_ok;
 }
 
@@ -274,30 +315,17 @@ oms_status_enu_t oms::Model::list(const oms::ComRef& cref, char** contents)
 
   xmlStringWriter writer;
   pugi::xml_document doc;
-  pugi::xml_document ssvdoc;
 
+  Snapshot snapshot;
   // check for toplevelSystem or Model to update parameterbindings in ssd
   bool isTopSystemOrModel = false;
 
-  // generate XML declaration for ssv file
-  pugi::xml_node ssvDeclarationNode = ssvdoc.append_child(pugi::node_declaration);
-  ssvDeclarationNode.append_attribute("version") = "1.0";
-  ssvDeclarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node_parameterset = ssvdoc.append_child(oms::ssp::Version1_0::ssv::parameter_set);
-  node_parameterset.append_attribute("version") = "1.0";
-  node_parameterset.append_attribute("name") = "parameters";
-  pugi::xml_node node_parameters = node_parameterset.append_child(oms::ssp::Version1_0::ssv::parameters);
-
   // list model
   if (cref.isEmpty())
   {
     isTopSystemOrModel = true;
-    pugi::xml_node node = doc.append_child(oms::ssp::Draft20180219::ssd::system_structure_description);
-    exportToSSD(node, node_parameters);
-    // update parameterBindings in ssd
-    pugi::xml_node system_node = node.child(oms::ssp::Draft20180219::ssd::system);
-    updateParameterBindingsToSSD(system_node, node_parameters, isTopSystemOrModel);
+    exportToSSD(snapshot);
+    doc.append_copy(snapshot.getResourceNode("SystemStructure.ssd"));
   }
   else
   {
@@ -312,10 +340,11 @@ oms_status_enu_t oms::Model::list(const oms::ComRef& cref, char** contents)
 
     if (subsystem)
     {
-      pugi::xml_node node = doc.append_child(oms::ssp::Draft20180219::ssd::system);
-      subsystem->exportToSSD(node, node_parameters);
-      // update parameterBindings in ssd
-      updateParameterBindingsToSSD(node, node_parameters, isTopSystemOrModel);
+      pugi::xml_node ssdNode = snapshot.getTemplateResourceNodeSSD("SystemStructure.ssd", this->getCref());
+      pugi::xml_node system_node = ssdNode.append_child(oms::ssp::Draft20180219::ssd::system);
+
+      subsystem->exportToSSD(system_node, snapshot);
+      doc.append_copy(snapshot.getResourceNode("SystemStructure.ssd").first_child());
     }
     else
     {
@@ -324,81 +353,36 @@ oms_status_enu_t oms::Model::list(const oms::ComRef& cref, char** contents)
       if (!component)
         return logError("error");
 
-      pugi::xml_node node = doc.append_child(oms::ssp::Draft20180219::ssd::system);
-      component->exportToSSD(node, node_parameters);
+      pugi::xml_node ssdNode = snapshot.getTemplateResourceNodeSSD("SystemStructure.ssd", this->getCref());
+      pugi::xml_node system_node = ssdNode.append_child(oms::ssp::Draft20180219::ssd::system);
+
+      component->exportToSSD(system_node, snapshot);
+      doc.append_copy(snapshot.getResourceNode("SystemStructure.ssd").first_child());
     }
   }
 
   doc.save(writer);
-  *contents = (char*) malloc(strlen(writer.result.c_str()) + 1);
-  strcpy(*contents, writer.result.c_str());
+  *contents = mallocAndCopyString(writer.result);
+  if (!*contents)
+    return oms_status_fatal;
+
   return oms_status_ok;
 }
 
 oms_status_enu_t oms::Model::exportSnapshot(const oms::ComRef& cref, char** contents)
 {
-  // only top level model is allowed
-  if (!cref.isEmpty())
-  {
-    //return logError("only top level model is allowed, unknown model: " + std::string(cref));
-    return logError("\"" + std::string(getCref()+std::string(cref)) + "\" is not a top level model");
-  }
+  Snapshot snapshot;
 
-  struct xmlStringWriter : pugi::xml_writer
-  {
-    std::string result;
-    virtual void write(const void* data, size_t size)
-    {
-      result += std::string(static_cast<const char*>(data), size);
-    }
-  };
+  exportToSSD(snapshot);
+  exportSignalFilter(snapshot);
 
-  xmlStringWriter writer;
-  pugi::xml_document doc;
-  pugi::xml_document ssvdoc;
+  if (cref.isEmpty())
+    return snapshot.writeDocument(contents);
 
-  // generate XML declaration for ssv file
-  pugi::xml_node ssvDeclarationNode = ssvdoc.append_child(pugi::node_declaration);
-  ssvDeclarationNode.append_attribute("version") = "1.0";
-  ssvDeclarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node_parameterset = ssvdoc.append_child(oms::ssp::Version1_0::ssv::parameter_set);
-  node_parameterset.append_attribute("xmlns:ssc") = "http://ssp-standard.org/SSP1/SystemStructureCommon";
-  node_parameterset.append_attribute("xmlns:ssv") = "http://ssp-standard.org/SSP1/SystemStructureParameterValues";
-  node_parameterset.append_attribute("version") = "1.0";
-  node_parameterset.append_attribute("name") = "parameters";
-  pugi::xml_node node_parameters = node_parameterset.append_child(oms::ssp::Version1_0::ssv::parameters);
-
-  // list model
-  pugi::xml_node snapshotnode = doc.append_child(oms::ssp::Version1_0::snap_shot);
-  pugi::xml_node ssdfilenode = snapshotnode.append_child(oms::ssp::Version1_0::ssd_file);
-  ssdfilenode.append_attribute("name") = "SystemStructure.ssd";
-
-  pugi::xml_node node = ssdfilenode.append_child(oms::ssp::Draft20180219::ssd::system_structure_description);
-  exportToSSD(node, node_parameters);
-
-  // update ssv file if exist
-  if (!Flags::ExportParametersInline())
-  {
-    // update parameterBindings in ssd
-    pugi::xml_node system_node = node.child(oms::ssp::Draft20180219::ssd::system);
-    updateParameterBindingsToSSD(system_node, node_parameters, true);
-
-    // update <oms:ssv_file> after </oms:ssd_file>
-    pugi::xml_node last = doc.last_child();
-    pugi::xml_node ssvfilenode  = last.append_child(oms::ssp::Version1_0::ssv_file);
-    std::string ssvFilePath = "resources/" + std::string(this->getCref()) + ".ssv";
-    ssvfilenode.append_attribute("name") = ssvFilePath.c_str();
-    // dump all the ssv file contents
-    ssvfilenode.append_copy(node_parameterset);
-    // TODO ssm file
-  }
-
-  doc.save(writer);
-  *contents = (char*) malloc(strlen(writer.result.c_str()) + 1);
-  strcpy(*contents, writer.result.c_str());
-
-  return oms_status_ok;
+  // query for partial snapshot
+  Snapshot partialSnapshot(true);
+  snapshot.exportPartialSnapshot(cref, partialSnapshot);
+  return partialSnapshot.writeDocument(contents);
 }
 
 oms_status_enu_t oms::Model::exportSSVTemplate(const oms::ComRef& cref, const std::string& filename)
@@ -407,6 +391,10 @@ oms_status_enu_t oms::Model::exportSSVTemplate(const oms::ComRef& cref, const st
   oms::ComRef front = tail.pop_front();
 
   pugi::xml_document ssvdoc;
+  // generate XML declaration for ssm file
+  pugi::xml_node ssvDeclarationNode = ssvdoc.append_child(pugi::node_declaration);
+  ssvDeclarationNode.append_attribute("version") = "1.0";
+  ssvDeclarationNode.append_attribute("encoding") = "UTF-8";
 
   std::string extension = "";
   if (filename.length() > 4)
@@ -415,24 +403,14 @@ oms_status_enu_t oms::Model::exportSSVTemplate(const oms::ComRef& cref, const st
   if (extension != ".ssv")
     return logError("filename extension must be \".ssv\"; no other formats are supported");
 
-  // generate XML declaration for ssv file
-  pugi::xml_node ssvDeclarationNode = ssvdoc.append_child(pugi::node_declaration);
-  ssvDeclarationNode.append_attribute("version") = "1.0";
-  ssvDeclarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node_parameterset = ssvdoc.append_child(oms::ssp::Version1_0::ssv::parameter_set);
-  node_parameterset.append_attribute("xmlns:ssc") = "http://ssp-standard.org/SSP1/SystemStructureCommon";
-  node_parameterset.append_attribute("xmlns:ssv") = "http://ssp-standard.org/SSP1/SystemStructureParameterValues";
-
-  node_parameterset.append_attribute("version") = "1.0";
-  node_parameterset.append_attribute("name") = "modelDescriptionStartValues";
-  pugi::xml_node node_parameters = node_parameterset.append_child(oms::ssp::Version1_0::ssv::parameters);
+  Snapshot snapshot;
+  pugi::xml_node ssvNode = snapshot.getTemplateResourceNodeSSV("template.ssv", "modelDescriptionStartValues");
 
   for (const auto& component : system->getComponents())
   {
     if(component.first == tail || cref.isEmpty()) // allow querying single component or whole model
     {
-      if (oms_status_ok != component.second->exportToSSVTemplate(node_parameters))
+      if (oms_status_ok != component.second->exportToSSVTemplate(ssvNode))
       {
         return logError("export of ssv template failed for component " + std::string(system->getFullCref()+std::string(component.first)));
       }
@@ -440,8 +418,9 @@ oms_status_enu_t oms::Model::exportSSVTemplate(const oms::ComRef& cref, const st
   }
 
   //ssvdoc.save(std::cout);
+  ssvdoc.append_copy(snapshot.getResourceNode("template.ssv"));
 
-  if (!ssvdoc.save_file(filename.c_str()))
+  if (!ssvdoc.save_file(filename.c_str(), "  ", pugi::format_indent|pugi::format_indent_attributes, pugi::encoding_utf8))
     return logError("failed to export  \"" + filename + "\" (for model \"" + std::string(this->getCref()) + "\")");
 
   return oms_status_ok;
@@ -453,6 +432,10 @@ oms_status_enu_t oms::Model::exportSSMTemplate(const oms::ComRef& cref, const st
   oms::ComRef front = tail.pop_front();
 
   pugi::xml_document ssmdoc;
+  // generate XML declaration for ssm file
+  pugi::xml_node ssmDeclarationNode = ssmdoc.append_child(pugi::node_declaration);
+  ssmDeclarationNode.append_attribute("version") = "1.0";
+  ssmDeclarationNode.append_attribute("encoding") = "UTF-8";
 
   std::string extension = "";
   if (filename.length() > 4)
@@ -461,22 +444,14 @@ oms_status_enu_t oms::Model::exportSSMTemplate(const oms::ComRef& cref, const st
   if (extension != ".ssm")
     return logError("filename extension must be \".ssm\"; no other formats are supported");
 
-  // generate XML declaration for ssv file
-  pugi::xml_node ssmDeclarationNode = ssmdoc.append_child(pugi::node_declaration);
-  ssmDeclarationNode.append_attribute("version") = "1.0";
-  ssmDeclarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node_parameterMapping = ssmdoc.append_child(oms::ssp::Version1_0::ssm::parameter_mapping);
-  node_parameterMapping.append_attribute("xmlns:ssc") = "http://ssp-standard.org/SSP1/SystemStructureCommon";
-  node_parameterMapping.append_attribute("xmlns:ssm") = "http://ssp-standard.org/SSP1/SystemStructureParameterMapping";
-
-  node_parameterMapping.append_attribute("version") = "1.0";
+  Snapshot snapshot;
+  pugi::xml_node ssmNode = snapshot.getTemplateResourceNodeSSM("template.ssm");
 
   for (const auto& component : system->getComponents())
   {
     if(component.first == tail || cref.isEmpty()) // allow querying single component or whole model
     {
-      if (oms_status_ok != component.second->exportToSSMTemplate(node_parameterMapping))
+      if (oms_status_ok != component.second->exportToSSMTemplate(ssmNode))
       {
         return logError("export of ssm template failed for component " + std::string(system->getFullCref()+std::string(component.first)));
       }
@@ -484,8 +459,9 @@ oms_status_enu_t oms::Model::exportSSMTemplate(const oms::ComRef& cref, const st
   }
 
   //ssmdoc.save(std::cout);
+  ssmdoc.append_copy(snapshot.getResourceNode("template.ssm"));
 
-  if (!ssmdoc.save_file(filename.c_str()))
+  if (!ssmdoc.save_file(filename.c_str(), "  ", pugi::format_indent|pugi::format_indent_attributes, pugi::encoding_utf8))
     return logError("failed to export  \"" + filename + "\" (for model \"" + std::string(this->getCref()) + "\")");
 
   return oms_status_ok;
@@ -497,12 +473,10 @@ oms_status_enu_t oms::Model::exportSSMTemplate(const oms::ComRef& cref, const st
  *     <ssd:ParameterBinding source="resources/import_export_parameters.ssv" />
  * </ssd:ParameterBindings>
  */
-oms_status_enu_t oms::Model::updateParameterBindingsToSSD(pugi::xml_node& node, pugi::xml_node& ssvNode, bool isTopSystemOrModel) const
+oms_status_enu_t oms::Model::updateParameterBindingsToSSD(pugi::xml_node& node, bool isTopSystemOrModel) const
 {
-  int parameterNodeCount = std::distance(ssvNode.begin(), ssvNode.end());
-
   // check parameter bindings exist and export to ssv file and also update the ssd file with parameterBindings at the top level
-  if (parameterNodeCount > 0 && isTopSystemOrModel)
+  if (isTopSystemOrModel)
   {
     // update the ssd with the top level parameterBindings (e.g)  <ParameterBinding source="resources/ControlledTemperature.ssv">
     for(pugi::xml_node_iterator it = node.begin(); it != node.end(); ++it)
@@ -521,6 +495,17 @@ oms_status_enu_t oms::Model::updateParameterBindingsToSSD(pugi::xml_node& node,
   return oms_status_ok;
 }
 
+oms_status_enu_t oms::Model::addResources(const oms::ComRef& cref)
+{
+  ComRef subCref(cref);
+  std::string fileName = "resources/" + subCref.pop_suffix();
+
+  if (system)
+    return system->addResources(subCref, fileName);
+
+  return oms_status_ok;
+}
+
 oms_status_enu_t oms::Model::addSystem(const oms::ComRef& cref, oms_system_enu_t type)
 {
   if (cref.isValidIdent() && !system)
@@ -547,26 +532,18 @@ oms_status_enu_t oms::Model::addSystem(const oms::ComRef& cref, oms_system_enu_t
   return logError("wrong input \"" + std::string(front) + "\" != \"" + std::string(system->getCref()) + "\"");
 }
 
-oms_status_enu_t oms::Model::exportToSSD(pugi::xml_node& node, pugi::xml_node& ssvNode) const
+oms_status_enu_t oms::Model::exportToSSD(Snapshot& snapshot) const
 {
-  node.append_attribute("xmlns:ssc") = "http://ssp-standard.org/SSP1/SystemStructureCommon";
-  node.append_attribute("xmlns:ssd") = "http://ssp-standard.org/SSP1/SystemStructureDescription";
-  node.append_attribute("xmlns:ssv") = "http://ssp-standard.org/SSP1/SystemStructureParameterValues";
-  node.append_attribute("xmlns:ssm") = "http://ssp-standard.org/SSP1/SystemStructureParameterMapping";
-  node.append_attribute("xmlns:ssb") = "http://ssp-standard.org/SSP1/SystemStructureSignalDictionary";
-  node.append_attribute("xmlns:oms") = "https://raw.githubusercontent.com/OpenModelica/OMSimulator/master/schema/oms.xsd";
-
-  node.append_attribute("name") = this->getCref().c_str();
-  node.append_attribute("version") = "1.0";
+  pugi::xml_node ssdNode = snapshot.getTemplateResourceNodeSSD("SystemStructure.ssd", this->getCref());
 
   if (system)
   {
-    pugi::xml_node system_node = node.append_child(oms::ssp::Draft20180219::ssd::system);
-    if (oms_status_ok != system->exportToSSD(system_node, ssvNode))
+    pugi::xml_node system_node = ssdNode.append_child(oms::ssp::Draft20180219::ssd::system);
+    if (oms_status_ok != system->exportToSSD(system_node, snapshot))
       return logError("export of system failed");
   }
 
-  pugi::xml_node default_experiment = node.append_child(oms::ssp::Draft20180219::ssd::default_experiment);
+  pugi::xml_node default_experiment = ssdNode.append_child(oms::ssp::Draft20180219::ssd::default_experiment);
   default_experiment.append_attribute("startTime") = std::to_string(startTime).c_str();
   default_experiment.append_attribute("stopTime") = std::to_string(stopTime).c_str();
 
@@ -581,21 +558,25 @@ oms_status_enu_t oms::Model::exportToSSD(pugi::xml_node& node, pugi::xml_node& s
   oms_simulation_information.append_attribute("resultFile") = resultFilename.c_str();
   oms_simulation_information.append_attribute("loggingInterval") = std::to_string(loggingInterval).c_str();
   oms_simulation_information.append_attribute("bufferSize") = std::to_string(bufferSize).c_str();
-  oms_simulation_information.append_attribute("signalFilter") = signalFilter.c_str();
+  oms_simulation_information.append_attribute("signalFilter") = signalFilterFilename.c_str();
 
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::importFromSSD(const pugi::xml_node& node)
+oms_status_enu_t oms::Model::importFromSnapshot(const Snapshot& snapshot)
 {
-  std::string sspVersion = node.attribute("version").as_string();
+  pugi::xml_node ssdNode = snapshot.getResourceNode("SystemStructure.ssd");
+  if (!ssdNode)
+    return logError("loading <oms:file> \"SystemStructure.ssd\" from <oms:snapshot> failed");
+
+  std::string sspVersion = ssdNode.attribute("version").as_string();
 
   if(sspVersion == "Draft20180219")
   {
     logWarning_deprecated;
   }
 
-  for(pugi::xml_node_iterator it = node.begin(); it != node.end(); ++it)
+  for(pugi::xml_node_iterator it = ssdNode.begin(); it != ssdNode.end(); ++it)
   {
     std::string name = it->name();
     if (name == oms::ssp::Draft20180219::ssd::system)
@@ -612,7 +593,7 @@ oms_status_enu_t oms::Model::importFromSSD(const pugi::xml_node& node)
       if (!system)
         return oms_status_error;
 
-      if (oms_status_ok != system->importFromSSD(*it, sspVersion))
+      if (oms_status_ok != system->importFromSnapshot(*it, sspVersion, snapshot))
         return oms_status_error;
     }
     else if (name == oms::ssp::Draft20180219::ssd::default_experiment)
@@ -650,7 +631,9 @@ oms_status_enu_t oms::Model::importFromSSD(const pugi::xml_node& node)
             resultFilename = itAnnotations->attribute("resultFile").as_string();
             loggingInterval = itAnnotations->attribute("loggingInterval").as_double();
             bufferSize = itAnnotations->attribute("bufferSize").as_int();
-            setSignalFilter(itAnnotations->attribute("signalFilter").as_string());
+
+            if (oms_status_ok == importSignalFilter(itAnnotations->attribute("signalFilter").as_string(), snapshot))
+              this->signalFilterFilename = itAnnotations->attribute("signalFilter").as_string();
           }
         }
       }
@@ -738,8 +721,8 @@ oms_system_enu_t oms::Model::getSystemTypeHelper(const pugi::xml_node& node, con
 
 oms_status_enu_t oms::Model::exportToFile(const std::string& filename) const
 {
-  pugi::xml_document doc;
-  pugi::xml_document ssvdoc;
+
+  Snapshot snapshot;
 
   std::string extension = "";
   if (filename.length() > 4)
@@ -748,60 +731,8 @@ oms_status_enu_t oms::Model::exportToFile(const std::string& filename) const
   if (extension != ".ssp")
     return logError("filename extension must be \".ssp\"; no other formats are supported");
 
-  // generate XML declaration
-  pugi::xml_node declarationNode = doc.append_child(pugi::node_declaration);
-  declarationNode.append_attribute("version") = "1.0";
-  declarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node = doc.append_child(oms::ssp::Draft20180219::ssd::system_structure_description);
-
-  // generate XML declaration for ssv file
-  pugi::xml_node ssvDeclarationNode = ssvdoc.append_child(pugi::node_declaration);
-  ssvDeclarationNode.append_attribute("version") = "1.0";
-  ssvDeclarationNode.append_attribute("encoding") = "UTF-8";
-
-  pugi::xml_node node_parameterset = ssvdoc.append_child(oms::ssp::Version1_0::ssv::parameter_set);
-  node_parameterset.append_attribute("xmlns:ssc") = "http://ssp-standard.org/SSP1/SystemStructureCommon";
-  node_parameterset.append_attribute("xmlns:ssv") = "http://ssp-standard.org/SSP1/SystemStructureParameterValues";
-
-  node_parameterset.append_attribute("version") = "1.0";
-  node_parameterset.append_attribute("name") = "parameters";
-  pugi::xml_node node_parameters = node_parameterset.append_child(oms::ssp::Version1_0::ssv::parameters);
-
-  exportToSSD(node, node_parameters);
-
-  filesystem::path ssdPath = filesystem::path(tempDir) / "SystemStructure.ssd";
-
-  // check for parameter-bindings are defined, (i.e) count the child nodes node_parameters in ssvdoc
-  int parameterNodeCount = std::distance(node_parameters.begin(), node_parameters.end());
-  std::string ssvFileName = "";
-
-  // check parameter bindings exist and export to ssv file and also update the ssd file with parameterBindings at the top level
-  if (parameterNodeCount > 0)
-  {
-    ssvFileName = "resources/" + std::string(this->getCref()) + ".ssv";
-    filesystem::path ssvPath = filesystem::path(tempDir) /  ssvFileName;
-    //std::cout << "\n ssvPath  : " << ssvPath << " filename : " << ssvFileName;
-    ssvdoc.save_file(ssvPath.string().c_str());
-
-    // update the ssd with the top level parameterBindings (e.g)  <ParameterBinding source="resources/ControlledTemperature.ssv">
-    for(pugi::xml_node_iterator it = node.begin(); it != node.end(); ++it)
-    {
-      pugi::xml_node node_elements = it->child(oms::ssp::Draft20180219::ssd::elements);
-      if (node_elements) // insert the parameter bindings before <ssd:Elements>
-      {
-        pugi::xml_node node_parameters_bindings = it->insert_child_before(oms::ssp::Version1_0::ssd::parameter_bindings, node_elements);
-        pugi::xml_node node_parameter_binding  = node_parameters_bindings.append_child(oms::ssp::Version1_0::ssd::parameter_binding);
-        node_parameter_binding.append_attribute("source") = ssvFileName.c_str();
-        break;
-      }
-    }
-  }
-
-  //doc.save(std::cout);
-
-  if (!doc.save_file(ssdPath.string().c_str()))
-    return logError("failed to export \"" + ssdPath.string() + "\" (for model \"" + std::string(this->getCref()) + "\")");
+  exportToSSD(snapshot);
+  exportSignalFilter(snapshot);
 
   // Usage: minizip [-o] [-a] [-0 to -9] [-p password] [-j] file.zip [files_to_add]
   //        -o  Overwrite existing file.zip
@@ -812,12 +743,7 @@ oms_status_enu_t oms::Model::exportToFile(const std::string& filename) const
   //        -j  exclude path. store only the file name
 
   std::vector<std::string> resources;
-  if (!ssvFileName.empty())
-  {
-    resources.push_back(ssvFileName);
-  }
-  if (oms_status_ok != getAllResources(resources))
-    return logError("failed to gather all resources");
+  writeAllResourcesToFilesystem(resources, snapshot);
 
   std::string cd = Scope::GetInstance().getWorkingDirectory();
   Scope::GetInstance().setWorkingDirectory(tempDir);
@@ -840,18 +766,24 @@ oms_status_enu_t oms::Model::exportToFile(const std::string& filename) const
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::getAllResources(std::vector<std::string>& resources) const
+void oms::Model::writeAllResourcesToFilesystem(std::vector<std::string>& resources, Snapshot& snapshot) const
 {
-  resources.push_back("SystemStructure.ssd");
+  // get all files from snapshot and save the document (eg.) ssd, ssv and signalFilter
+  snapshot.getResources(resources);
+  for (auto const &filename : resources)
+  {
+    if (oms_status_ok != snapshot.writeResourceNode(filename, filesystem::path(tempDir)))
+      logError("failed to export \"" + filename + " to directory " + tempDir);
+  }
+
   if (system)
-    return system->getAllResources(resources);
-  return oms_status_ok;
+    system->getAllResources(resources);
 }
 
 oms_status_enu_t oms::Model::setStartTime(double value)
 {
   if (!validState(oms_modelState_virgin|oms_modelState_enterInstantiation|oms_modelState_instantiated))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   startTime = value;
   return oms_status_ok;
@@ -860,16 +792,23 @@ oms_status_enu_t oms::Model::setStartTime(double value)
 oms_status_enu_t oms::Model::setStopTime(double value)
 {
   if (!validState(oms_modelState_virgin|oms_modelState_enterInstantiation|oms_modelState_instantiated))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   stopTime = value;
   return oms_status_ok;
 }
 
+double oms::Model::getTime() const
+{
+  if (system && validState(oms_modelState_simulation))
+    return system->getTime();
+  return getStartTime();
+}
+
 oms_status_enu_t oms::Model::instantiate()
 {
   if (!validState(oms_modelState_virgin|oms_modelState_enterInstantiation))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   if (!system)
     return logError("Model doesn't contain a system");
@@ -896,7 +835,7 @@ oms_status_enu_t oms::Model::instantiate()
 oms_status_enu_t oms::Model::initialize()
 {
   if (!validState(oms_modelState_instantiated))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   if (!system)
     return logError("Model doesn't contain a system");
@@ -905,7 +844,6 @@ oms_status_enu_t oms::Model::initialize()
   clock.reset();
   clock.tic();
 
-  cancelSim = false;
   lastEmit = startTime;
 
   if (!resultFilename.empty())
@@ -967,25 +905,13 @@ oms_status_enu_t oms::Model::initialize()
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::simulate_asynchronous(void (*cb)(const char* cref, double time, oms_status_enu_t status))
-{
-  if (!validState(oms_modelState_simulation))
-    return logError_ModelInWrongState(this);
-
-  if (!system)
-    return logError("Model doesn't contain a system");
-
-  std::thread([=]{system->stepUntil(stopTime, cb);}).detach();
-  return oms_status_pending;
-}
-
 oms_status_enu_t oms::Model::simulate()
 {
   clock.tic();
   if (!validState(oms_modelState_simulation))
   {
     clock.toc();
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
   }
 
   if (!system)
@@ -994,19 +920,39 @@ oms_status_enu_t oms::Model::simulate()
     return logError("Model doesn't contain a system");
   }
 
-  oms_status_enu_t status = system->stepUntil(stopTime, NULL);
+  oms_status_enu_t status = system->stepUntil(stopTime);
   emit(stopTime, true);
   clock.toc();
   return status;
 }
 
+oms_status_enu_t oms::Model::doStep()
+{
+  clock.tic();
+  if (!validState(oms_modelState_simulation))
+  {
+    clock.toc();
+    return logError_ModelInWrongState(getCref());
+  }
+
+  if (!system)
+  {
+    clock.toc();
+    return logError("Model doesn't contain a system");
+  }
+
+  oms_status_enu_t status = system->doStep();
+  clock.toc();
+  return status;
+}
+
 oms_status_enu_t oms::Model::stepUntil(double stopTime)
 {
   clock.tic();
   if (!validState(oms_modelState_simulation))
   {
     clock.toc();
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
   }
 
   if (!system)
@@ -1015,7 +961,7 @@ oms_status_enu_t oms::Model::stepUntil(double stopTime)
     return logError("Model doesn't contain a system");
   }
 
-  oms_status_enu_t status = system->stepUntil(stopTime, NULL);
+  oms_status_enu_t status = system->stepUntil(stopTime);
   emit(stopTime, true);
   clock.toc();
   return status;
@@ -1027,7 +973,7 @@ oms_status_enu_t oms::Model::terminate()
     return oms_status_ok;
 
   if (validState(oms_modelState_enterInstantiation))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   if (!system)
     return logError("Model doesn't contain a system");
@@ -1054,7 +1000,7 @@ oms_status_enu_t oms::Model::terminate()
 oms_status_enu_t oms::Model::reset()
 {
   if (!validState(oms_modelState_simulation))
-    return logError_ModelInWrongState(this);
+    return logError_ModelInWrongState(getCref());
 
   if (!system)
     return logError("Model doesn't contain a system");
@@ -1165,32 +1111,6 @@ oms_status_enu_t oms::Model::getResultFile(char** filename, int* bufferSize)
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::getSignalFilter(char** regex)
-{
-  *regex = (char*)this->signalFilter.c_str();
-
-  return oms_status_ok;
-}
-
-oms_status_enu_t oms::Model::setSignalFilter(const std::string& regex)
-{
-  // If regex is empty then all signals will be exported
-  if (regex.empty() || regex == ".*")
-    this->signalFilter = ".*";
-  else
-  {
-    if (oms_status_ok != removeSignalsFromResults(".*"))
-      return oms_status_error;
-
-    this->signalFilter = regex;
-  }
-
-  if (oms_status_ok != system->addSignalsToResults(this->signalFilter.c_str()))
-    return oms_status_error;
-
-  return oms_status_ok;
-}
-
 oms_status_enu_t oms::Model::addSignalsToResults(const char* regex)
 {
   if (system)
@@ -1207,11 +1127,44 @@ oms_status_enu_t oms::Model::removeSignalsFromResults(const char* regex)
   return oms_status_ok;
 }
 
-oms_status_enu_t oms::Model::cancelSimulation_asynchronous()
+void oms::Model::exportSignalFilter(Snapshot& snapshot) const
 {
-  if (!validState(oms_modelState_simulation))
-    return logError_ModelInWrongState(this);
+  if (!system)
+    return;
+
+  pugi::xml_node oms_signalFilter = snapshot.getTemplateResourceNodeSignalFilter(signalFilterFilename);
+
+  std::vector<oms::Connector> filteredSignals;
+  system->getFilteredSignals(filteredSignals);
+
+  for (auto const& signal : filteredSignals)
+  {
+    pugi::xml_node oms_variable = oms_signalFilter.append_child(oms::ssp::Version1_0::oms_Variable);
+    oms_variable.append_attribute("name") = signal.getFullName().c_str();
+    oms_variable.append_attribute("type") = signal.getTypeString().c_str();
+    oms_variable.append_attribute("kind") = signal.getCausalityString().c_str();
+  }
+}
+
+oms_status_enu_t oms::Model::importSignalFilter(const std::string& filename, const Snapshot& snapshot)
+{
+  if (".*" == filename) // avoid error messages for older ssp files
+  {
+    addSignalsToResults(".*");
+    return oms_status_warning;
+  }
+
+  pugi::xml_node oms_signalfilter = snapshot.getResourceNode(filename);
+
+  if (!oms_signalfilter)
+    return oms_status_error;
+
+  removeSignalsFromResults(".*"); // disable all signals
+  for (pugi::xml_node_iterator it = oms_signalfilter.begin(); it != oms_signalfilter.end(); ++it)
+  {
+    if (std::string(it->name()) == oms::ssp::Version1_0::oms_Variable)
+      addSignalsToResults(it->attribute("name").as_string());
+  }
 
-  cancelSim = true;
   return oms_status_ok;
 }
